Package: fiber
Type: Package
Title: S7 Data Structures for Diffusion MRI Tractography
Version: 0.1.2
Authors@R: c(
    person(
        given = "Aymeric",
        family = "Stamm",
        role = c("aut", "cre"),
        email = "aymeric.stamm@cnrs.fr",
        comment = c(ORCID = "0000-0002-8725-3654")
    ))
Description: Provides three S7 classes — streamline, bundle, and bundle_set —
    for representing diffusion MRI tractography data in R, together with a
    concise set of methods for computing shape descriptors (arc-length,
    curvature, torsion, sinuosity), the Hausdorff distance between streamlines,
    arc-length reparametrization of streamlines and bundles onto uniform grids,
    combination of streamlines or bundles into a single bundle, combination of
    bundles from multiple subjects or sessions into a bundle_set, and coercion
    to and from the dwiFiber S4 class of the 'dti' package. See Dell'Acqua, F.,
    Descoteaux, M. and Leemans, A. (2024) "Handbook of Diffusion MR Tractography"
    <doi:10.1016/C2018-0-02520-7> for more about the mathematical and computational
    underpinnings of diffusion MRI tractography.
License: MIT + file LICENSE
Encoding: UTF-8
URL: https://github.com/tractoverse/fiber,
        https://tractoverse.github.io/fiber/
BugReports: https://github.com/tractoverse/fiber/issues
Imports: cli, methods, S7
Config/roxygen2/version: 8.0.0
Config/roxygen2/markdown: TRUE
Suggests: dti, tinytest
Collate: 'bundle.R' 'streamline.R' 'coerce.R' 'cpp11.R'
        'fiber-package.R' 'parametrize.R' 'shape.R' 'utils.R' 'zzz.R'
LinkingTo: cpp11
NeedsCompilation: yes
Packaged: 2026-05-27 17:37:49 UTC; stamm-a
Author: Aymeric Stamm [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-8725-3654>)
Maintainer: Aymeric Stamm <aymeric.stamm@cnrs.fr>
Repository: CRAN
Date/Publication: 2026-05-30 13:40:15 UTC
Built: R 4.5.2; aarch64-apple-darwin20; 2026-05-30 17:19:20 UTC; unix
Archs: fiber.so.dSYM

Package: fiber
Type: Package
Title: S7 Data Structures for Diffusion MRI Tractography
Version: 0.1.2
Authors@R: c(
    person(
        given = "Aymeric",
        family = "Stamm",
        role = c("aut", "cre"),
        email = "aymeric.stamm@cnrs.fr",
        comment = c(ORCID = "0000-0002-8725-3654")
    ))
Description: Provides three S7 classes — streamline, bundle, and bundle_set —
    for representing diffusion MRI tractography data in R, together with a
    concise set of methods for computing shape descriptors (arc-length,
    curvature, torsion, sinuosity), the Hausdorff distance between streamlines,
    arc-length reparametrization of streamlines and bundles onto uniform grids,
    combination of streamlines or bundles into a single bundle, combination of
    bundles from multiple subjects or sessions into a bundle_set, and coercion
    to and from the dwiFiber S4 class of the 'dti' package. See Dell'Acqua, F.,
    Descoteaux, M. and Leemans, A. (2024) "Handbook of Diffusion MR Tractography"
    <doi:10.1016/C2018-0-02520-7> for more about the mathematical and computational
    underpinnings of diffusion MRI tractography.
License: MIT + file LICENSE
Encoding: UTF-8
URL: https://github.com/tractoverse/fiber,
        https://tractoverse.github.io/fiber/
BugReports: https://github.com/tractoverse/fiber/issues
Imports: cli, methods, S7
Config/roxygen2/version: 8.0.0
Config/roxygen2/markdown: TRUE
Suggests: dti, tinytest
Collate: 'bundle.R' 'streamline.R' 'coerce.R' 'cpp11.R'
        'fiber-package.R' 'parametrize.R' 'shape.R' 'utils.R' 'zzz.R'
LinkingTo: cpp11
NeedsCompilation: yes
Packaged: 2026-05-27 17:37:49 UTC; stamm-a
Author: Aymeric Stamm [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-8725-3654>)
Maintainer: Aymeric Stamm <aymeric.stamm@cnrs.fr>
Repository: CRAN
Date/Publication: 2026-05-30 13:40:15 UTC
Built: R 4.6.0; x86_64-apple-darwin20; 2026-05-30 16:59:09 UTC; unix
Archs: fiber.so.dSYM

# Generated by roxygen2: do not edit by hand

export(add_shape_descriptors)
export(as_bundle)
export(as_bundle_set)
export(as_dwifiber)
export(as_streamline)
export(bind_bundle_sets)
export(bind_bundles)
export(bundle)
export(bundle_set)
export(compute_hausdorff_distance)
export(get_curvature)
export(get_curvilinear_length)
export(get_euclidean_length)
export(get_sinuosity)
export(get_torsion)
export(is_bundle)
export(is_bundle_set)
export(is_streamline)
export(reparametrize)
export(streamline)
useDynLib(fiber, .registration = TRUE)

<!-- README.md is generated from README.qmd. Please edit that file -->

# fiber

<!-- badges: start -->

[![R-CMD-check](https://github.com/tractoverse/fiber/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/tractoverse/fiber/actions/workflows/R-CMD-check.yaml)
[![Codecov test
coverage](https://codecov.io/gh/tractoverse/fiber/graph/badge.svg)](https://app.codecov.io/gh/tractoverse/fiber)
[![pkgdown](https://github.com/tractoverse/fiber/actions/workflows/pkgdown.yaml/badge.svg)](https://github.com/tractoverse/fiber/actions/workflows/pkgdown.yaml)
[![CRAN
status](https://www.r-pkg.org/badges/version/fiber)](https://CRAN.R-project.org/package=fiber)
<!-- badges: end -->

## Overview

The [**fiber**](https://tractoverse.github.io/fiber/) package provides
three [S7](https://rconsortium.github.io/S7/) classes — `streamline`,
`bundle`, and `bundle_set` — for representing diffusion MRI tractography
data in R, together with a concise set of methods:

| Function | Description |
|----|----|
| `streamline()` | Construct a single streamline |
| `bundle()` | Construct a bundle of streamlines |
| `bundle_set()` | Construct a named set of bundles (multi-subject / multi-session) |
| `is_streamline()` | Test if an object is a `streamline` |
| `is_bundle()` | Test if an object is a `bundle` |
| `is_bundle_set()` | Test if an object is a `bundle_set` |
| `bind_bundles()` | Combine streamlines and/or bundles into a single bundle |
| `bind_bundle_sets()` | Combine named bundles and/or bundle sets into a single bundle set |
| `as_bundle_set()` | Coerce to `bundle_set` class |
| `reparametrize()` | Resample onto a uniform arc-length grid |
| `as_dwifiber()` | Coerce to `dwiFiber` class from [](https://cran.r-project.org/package=dti) |
| `as_streamline()` | Coerce to `streamline` class |
| `as_bundle()` | Coerce to `bundle` class |
| `get_euclidean_length()` | Straight-line distance between endpoints |
| `get_curvilinear_length()` | Total arc-length |
| `get_sinuosity()` | Curvilinear / Euclidean length ratio |
| `get_curvature()` | Curvature profile or summary scalar |
| `get_torsion()` | Torsion profile or summary scalar |
| `add_shape_descriptors()` | Add curvature, torsion, and/or sinuosity to `@point_data` and/or `@streamline_data` |
| `compute_hausdorff_distance()` | Symmetric Hausdorff distance |

### Classes

A `streamline` stores:

- `@points` — an $n \times 3$ numeric matrix with columns `"X"`, `"Y"`,
  `"Z"`.
- `@point_data` — a named list of per-point vectors of length $n$ (any
  type).
- `@streamline_data` — a named list of per-streamline scalars (any
  type).

A `bundle` stores:

- `@streamlines` — a list of `streamline` objects.
- `@bundle_data` — a named list of bundle-level metadata.

A `bundle_set` stores:

- `@bundles` — a *named* list of `bundle` objects (names typically
  encode subject or session IDs, e.g. `"sub-01"`).
- `@set_data` — a named list of set-level metadata.

## Installation

You can install the development version of **fiber** from
[GitHub](https://github.com/tractoverse/fiber) with:

``` r
# install.packages("pak")
pak::pak("tractoverse/fiber")
```

## Quick start

``` r
library(fiber)

# Build a helix streamline (50 points)
t <- seq(0, 2 * pi, length.out = 50)
sl <- streamline(
  points = cbind(X = cos(t), Y = sin(t), Z = t / (2 * pi))
)
sl
#> <streamline [50 pts]>

# Shape descriptors
get_curvilinear_length(sl)
#> [1] 6.358015
get_sinuosity(sl)
#> [1] 6.358015
head(get_curvature(sl))
#> [1] 1.537338e-05 1.625127e-02 4.592627e-02 8.636675e-02 1.352354e-01
#> [6] 1.906608e-01

# Bundle two streamlines
sl2 <- streamline(
  points = cbind(X = cos(t) * 1.1, Y = sin(t) * 1.1, Z = t / (2 * pi))
)
b <- bind_bundles(sl, sl2)
b
#> <bundle [2 streamlines | 50–50 pts/streamline]>

# Reparametrize to 20 points each
b20 <- reparametrize(b, n_points = 20L)
b20
#> <bundle [2 streamlines | 20–20 pts/streamline]>

# Hausdorff distance
compute_hausdorff_distance(sl, sl2)
#> [1] 0.1

# Multi-subject: collect bundles from two subjects into a bundle_set
b_sub01 <- bundle(
  streamlines = list(sl),
  bundle_data = list(subject = "sub-01")
)
b_sub02 <- bundle(
  streamlines = list(sl2),
  bundle_data = list(subject = "sub-02")
)
bs <- bind_bundle_sets("sub-01" = b_sub01, "sub-02" = b_sub02)
bs
#> <bundle_set [2 bundles | 1–1 streamlines/bundle]: sub-01, sub-02>
bs[["sub-01"]]
#> <bundle [1 streamlines | 50–50 pts/streamline] | bundle: subject>
```

# ---- internal helpers -------------------------------------------------------

# Null-coalescing operator
`%||%` <- function(x, y) if (is.null(x)) y else x

# Euclidean distance between two length-3 numeric vectors
.dist3 <- function(p, q) sqrt(sum((p - q)^2L))

# Interpolate a numeric vector y measured at abscissa s onto new abscissa s_new
.approx1 <- function(s, y, s_new) stats::approx(s, y, xout = s_new)$y

# Compute cumulative arc-length abscissa for an n x 3 coordinate matrix.
# Returns a numeric vector of length n with s[1] == 0.
.arc_length <- function(pts) {
  diffs <- diff(pts)
  seg_len <- sqrt(rowSums(diffs^2L))
  c(0, cumsum(seg_len))
}

#' @include streamline.R bundle.R
NULL

# ---- shape descriptors computers ---------------------------------------

#' Euclidean length of a streamline
#'
#' `get_euclidean_length()` is a function that computes the Euclidean
#' (straight-line) distance of a [streamline] object.
#'
#' @param x A [streamline] object.
#' @returns A non-negative numeric scalar.
#' @export
#' @examples
#' pts <- matrix(runif(30), ncol = 3)
#' colnames(pts) <- c("X", "Y", "Z")
#' sl <- streamline(points = pts)
#' get_euclidean_length(sl)
get_euclidean_length <- function(x) {
  if (!is_streamline(x)) {
    cli::cli_abort("{.arg x} must be a {.cls fiber::streamline} object.")
  }
  pts <- x@points
  n <- nrow(pts)
  .dist3(pts[1L, ], pts[n, ])
}

#' Curvilinear length of a streamline
#'
#' `get_curvilinear_length()` is a function that computes the total arc-length of a
#' [streamline] object as the sum of Euclidean segment lengths between
#' consecutive points.
#'
#' @param x A [streamline] object.
#' @returns A non-negative numeric scalar.
#' @export
#' @examples
#' pts <- matrix(runif(30), ncol = 3)
#' colnames(pts) <- c("X", "Y", "Z")
#' sl <- streamline(points = pts)
#' get_curvilinear_length(sl)
get_curvilinear_length <- function(x) {
  if (!is_streamline(x)) {
    cli::cli_abort("{.arg x} must be a {.cls fiber::streamline} object.")
  }
  pts <- x@points
  s <- .arc_length(pts)
  s[length(s)]
}

#' Sinuosity of a streamline
#'
#' `get_sinuosity()` is a function that computes the ratio of curvilinear
#' length to Euclidean length for a [streamline] object, with a value of
#' 1 indicating a perfectly straight streamline and larger values indicating
#' greater curviness.
#'
#' @param x A [streamline] object.
#' @returns A numeric scalar \eqn{\ge 1}.
#' @export
#' @examples
#' pts <- matrix(runif(30), ncol = 3)
#' colnames(pts) <- c("X", "Y", "Z")
#' sl <- streamline(points = pts)
#' get_sinuosity(sl)
get_sinuosity <- function(x) {
  if (!is_streamline(x)) {
    cli::cli_abort("{.arg x} must be a {.cls fiber::streamline} object.")
  }
  get_curvilinear_length(x) / get_euclidean_length(x)
}

#' Curvature of a streamline
#'
#' `get_curvature()` is function that computes the curvature of a
#' [streamline] object. The curvature \eqn{\kappa(s)} at each point
#' along the arc-length abscissa is computed using cubic smoothing
#' splines (3 degrees of freedom per component).
#'
#' @param x A [streamline] object.
#' @returns A non-negative numeric vector of length `x@n_points` giving the
#'   curvature \eqn{\kappa(s)} at each sampled point along the streamline.
#'   Higher values indicate sharper bending at that location.
#' @export
#' @examples
#' pts <- matrix(runif(30), ncol = 3)
#' colnames(pts) <- c("X", "Y", "Z")
#' sl <- streamline(points = pts)
#' get_curvature(sl)
get_curvature <- function(x) {
  if (!is_streamline(x)) {
    cli::cli_abort("{.arg x} must be a {.cls fiber::streamline} object.")
  }
  pts <- x@points
  s <- .arc_length(pts)

  ssx <- stats::smooth.spline(s, pts[, "X"], df = 3L)
  ssy <- stats::smooth.spline(s, pts[, "Y"], df = 3L)
  ssz <- stats::smooth.spline(s, pts[, "Z"], df = 3L)

  dx <- stats::predict(ssx, s, 1L)$y
  dy <- stats::predict(ssy, s, 1L)$y
  dz <- stats::predict(ssz, s, 1L)$y
  d2x <- stats::predict(ssx, s, 2L)$y
  d2y <- stats::predict(ssy, s, 2L)$y
  d2z <- stats::predict(ssz, s, 2L)$y

  num <- (d2z * dy - d2y * dz)^2L +
    (d2x * dz - d2z * dx)^2L +
    (d2y * dx - d2x * dy)^2L
  denom <- (dx^2L + dy^2L + dz^2L)^3L
  sqrt(num / denom)
}

#' Torsion of a streamline
#'
#' `get_torsion()` is function that computes the torsion of a
#' [streamline] object. The torsion \eqn{\tau(s)} at each point
#' along the arc-length abscissa is computed using cubic smoothing
#' splines (4 degrees of freedom per component).
#'
#' @param x A [streamline] object.
#' @returns A numeric vector of length `x@n_points` giving the torsion
#'   \eqn{\tau(s)} at each sampled point along the streamline. Positive values
#'   indicate right-handed twisting; negative values indicate left-handed
#'   twisting; zero indicates a planar curve at that location.
#' @export
#' @examples
#' pts <- matrix(runif(30), ncol = 3)
#' colnames(pts) <- c("X", "Y", "Z")
#' sl <- streamline(points = pts)
#' get_torsion(sl)
get_torsion <- function(x) {
  if (!is_streamline(x)) {
    cli::cli_abort("{.arg x} must be a {.cls fiber::streamline} object.")
  }
  pts <- x@points
  s <- .arc_length(pts)

  ssx <- stats::smooth.spline(s, pts[, "X"], df = 4L)
  ssy <- stats::smooth.spline(s, pts[, "Y"], df = 4L)
  ssz <- stats::smooth.spline(s, pts[, "Z"], df = 4L)

  dx <- stats::predict(ssx, s, 1L)$y
  dy <- stats::predict(ssy, s, 1L)$y
  dz <- stats::predict(ssz, s, 1L)$y
  d2x <- stats::predict(ssx, s, 2L)$y
  d2y <- stats::predict(ssy, s, 2L)$y
  d2z <- stats::predict(ssz, s, 2L)$y
  d3x <- stats::predict(ssx, s, 3L)$y
  d3y <- stats::predict(ssy, s, 3L)$y
  d3z <- stats::predict(ssz, s, 3L)$y

  num <- d3x *
    (dy * d2z - d2y * dz) +
    d3y * (d2x * dz - dx * d2z) +
    d3z * (dx * d2y - d2x * dy)
  denom <- (dy * d2z - d2y * dz)^2L +
    (d2x * dz - dx * d2z)^2L +
    (dx * d2y - d2x * dy)^2L
  num / denom
}

# ---- generic: add_shape_descriptors ----------------------------------------

#' Adds shape descriptors to a streamline or bundle
#'
#' @description
#' `add_shape_descriptors()` is an S7 generic that computes a number of shape
#' descriptors for each [streamline] object and stores them in the
#' `@streamline_data` or `@point_data` slots as appropriate, with methods
#' available for the following classes:
#'
#' `r doclisting::methods_list("add_shape_descriptors")`
#'
#' This function provides a convenient way to compute shape descriptors and
#' attach them to [streamline] or [bundle] objects. See the documentation for
#' each individual shape descriptor function (e.g. `get_euclidean_length()`,
#' `get_curvilinear_length()`, `get_sinuosity()`, `get_curvature()`,
#' `get_torsion()`) for more details on how each descriptor is computed.
#'
#' @param x A [streamline] or [bundle] object.
#' @param descriptors A character vector of shape descriptors to add. Defaults
#' to all available descriptors: `c("euclidean_length", "curvilinear_length",
#' "sinuosity", "curvature", "torsion")`.
#' @returns An object of the same class as `x` with the specified shape descriptors
#' added to the `@streamline_data` or `@point_data` slots of each streamline.
#' @export
#' @examples
#' # add multiple shape descriptors to a single streamline
#' pts <- matrix(runif(30), ncol = 3)
#' colnames(pts) <- c("X", "Y", "Z")
#' sl <- streamline(points = pts)
#' sl <- add_shape_descriptors(
#'   sl,
#'   descriptors = c("euclidean_length", "curvilinear_length", "sinuosity")
#' )
#' # add multiple shape descriptors to a bundle
#' sl1 <- streamline(points = pts)
#' pts2 <- matrix(runif(60), ncol = 3)
#' colnames(pts2) <- c("X", "Y", "Z")
#' sl2 <- streamline(points = pts2)
#' b <- bundle(streamlines = list(sl1, sl2))
#' b <- add_shape_descriptors(
#'   b,
#'   descriptors = c("euclidean_length", "curvilinear_length", "sinuosity")
#' )
add_shape_descriptors <- S7::new_generic(
  "add_shape_descriptors",
  "x",
  function(
    x,
    descriptors = c(
      "euclidean_length",
      "curvilinear_length",
      "sinuosity",
      "curvature",
      "torsion"
    )
  ) {
    S7::S7_dispatch()
  }
)

# --- method: streamline -----------------------------------------------------

#' [add_shape_descriptors()] method for `streamline` objects
#'
#' Adds multiple shape descriptors to a single [streamline] object.
#'
#' @param x A [streamline] object.
#' @inheritParams add_shape_descriptors
#' @returns A [streamline] with the specified shape descriptors added to the
#' `@streamline_data` or `@point_data` slots as appropriate.
#' @seealso [add_shape_descriptors()]
#' @name add_shape_descriptors-fiber-streamline-method
#' @aliases add_shape_descriptors,fiber::streamline-method
#' @usage NULL
S7::method(add_shape_descriptors, streamline) <- function(
  x,
  descriptors = c(
    "euclidean_length",
    "curvilinear_length",
    "sinuosity",
    "curvature",
    "torsion"
  )
) {
  for (desc in descriptors) {
    if (desc == "euclidean_length") {
      x@streamline_data$euclidean_length <- get_euclidean_length(x)
    } else if (desc == "curvilinear_length") {
      x@streamline_data$curvilinear_length <- get_curvilinear_length(x)
    } else if (desc == "sinuosity") {
      x@streamline_data$sinuosity <- get_sinuosity(x)
    } else if (desc == "curvature") {
      x@point_data$curvature <- get_curvature(x)
    } else if (desc == "torsion") {
      x@point_data$torsion <- get_torsion(x)
    } else {
      warning(
        cli::format_inline(
          "Unknown shape descriptor: {.val {desc}}. Currently available ",
          "descriptors are: {.val euclidean_length}, {.val curvilinear_length}, ",
          "{.val sinuosity}, {.val curvature}, and {.val torsion}."
        ),
        call. = FALSE
      )
    }
  }
  x
}

# --- method: bundle ---------------------------------------------------------

#' [add_shape_descriptors()] method for `bundle` objects
#'
#' Adds multiple shape descriptors to every [streamline] inside a [bundle].
#'
#' @param x A [bundle] object.
#' @inheritParams add_shape_descriptors
#' @returns A [bundle] with the specified shape descriptors added to the
#' `@streamline_data` or `@point_data` slots of each streamline as appropriate.
#' @seealso [add_shape_descriptors()]
#' @name add_shape_descriptors-fiber-bundle-method
#' @aliases add_shape_descriptors,fiber::bundle-method
#' @usage NULL
S7::method(add_shape_descriptors, bundle) <- function(
  x,
  descriptors = c(
    "euclidean_length",
    "curvilinear_length",
    "sinuosity",
    "curvature",
    "torsion"
  )
) {
  x@streamlines <- lapply(
    x@streamlines,
    add_shape_descriptors,
    descriptors = descriptors
  )
  x
}

# ---- generic: compute_hausdorff_distance ---------------------------------

#' Computes the Hausdorff distance between streamlines
#'
#' @description
#' `compute_hausdorff_distance()` is an S7 generic that computes the symmetric
#' Hausdorff distance between [streamline] objects based on their 3-D
#' coordinate matrices, with methods available for the following classes:
#'
#' `r doclisting::methods_list("compute_hausdorff_distance")`
#'
#' The four dispatch cases are:
#'
#' - **`streamline` + `streamline`**: returns a single numeric scalar — the
#'   symmetric Hausdorff distance between the two streamlines.
#' - **`bundle` + missing**: returns a symmetric numeric distance matrix of
#'   dimension \eqn{n \times n}, where \eqn{n} is the number of streamlines in
#'   the bundle, giving all pairwise Hausdorff distances.
#' - **`bundle` + `streamline`**: returns a numeric vector of length \eqn{n}
#'   giving the Hausdorff distance from `y` to each streamline in `x`.
#' - **`bundle` + `bundle`**: returns a symmetric numeric distance matrix of
#'   dimension \eqn{(n_x + n_y) \times (n_x + n_y)}, treating the
#'   concatenation of all streamlines from `x` and `y` as one collection.
#'
#' @param x A [streamline] or [bundle] object.
#' @param y A [streamline] or [bundle] object, or `NULL` (default). When `NULL`
#'   and `x` is a [bundle], the pairwise distance matrix within `x` is
#'   returned.
#' @returns
#'   - A non-negative numeric scalar when both `x` and `y` are [streamline]s.
#'   - A [`dist`][stats::dist] object of size \eqn{n} when `x` is a [bundle]
#'     and `y` is `NULL` or a [bundle] (use [as.matrix()] to expand to a full
#'     \eqn{n \times n} matrix).
#'   - A numeric vector of length \eqn{n} when `x` is a [bundle] and `y` is
#'     a [streamline].
#' @export
#' @examples
#' pts1 <- matrix(runif(30), ncol = 3)
#' colnames(pts1) <- c("X", "Y", "Z")
#' sl1 <- streamline(points = pts1)
#' pts2 <- matrix(runif(30), ncol = 3)
#' colnames(pts2) <- c("X", "Y", "Z")
#' sl2 <- streamline(points = pts2)
#'
#' # streamline x streamline -> scalar
#' compute_hausdorff_distance(sl1, sl2)
#'
#' # bundle x missing -> pairwise dist object
#' b <- bundle(streamlines = list(sl1, sl2))
#' compute_hausdorff_distance(b)
#' as.matrix(compute_hausdorff_distance(b))
#'
#' # bundle x streamline -> vector
#' compute_hausdorff_distance(b, sl1)
#'
#' # bundle x bundle -> combined pairwise matrix
#' b2 <- bundle(streamlines = list(sl2))
#' compute_hausdorff_distance(b, b2)
compute_hausdorff_distance <- S7::new_generic(
  "compute_hausdorff_distance",
  "x",
  function(x, y = NULL) {
    S7::S7_dispatch()
  }
)

# --- method: streamline x streamline ----------------------------------------

#' [compute_hausdorff_distance()] method for two `streamline` objects
#'
#' @param x A [streamline] object.
#' @param y A [streamline] object.
#' @returns A non-negative numeric scalar equal to
#'   \eqn{\max(d_H(x \to y),\, d_H(y \to x))}, where
#'   \eqn{d_H(A \to B) = \max_{a \in A} \min_{b \in B} \|a - b\|_2} is the
#'   directed Hausdorff distance. The core computation is performed in C++
#'   via `hausdorff_distance_cpp()`.
#' @seealso [compute_hausdorff_distance()]
#' @name compute_hausdorff_distance-fiber-streamline-method
#' @aliases compute_hausdorff_distance,fiber::streamline-method
#' @usage NULL
#' @examples
#' pts1 <- matrix(runif(30), ncol = 3)
#' colnames(pts1) <- c("X", "Y", "Z")
#' pts2 <- matrix(runif(30), ncol = 3)
#' colnames(pts2) <- c("X", "Y", "Z")
#' sl1 <- streamline(points = pts1)
#' sl2 <- streamline(points = pts2)
#' compute_hausdorff_distance(sl1, sl2)
S7::method(compute_hausdorff_distance, streamline) <- function(x, y = NULL) {
  if (!is_streamline(y)) {
    cli::cli_abort(c(
      "{.arg y} must be a {.cls fiber::streamline} object when {.arg x} is a
      {.cls fiber::streamline}.",
      "i" = "To compute pairwise distances within a bundle, pass a
      {.cls fiber::bundle} as {.arg x} and omit {.arg y}."
    ))
  }
  xyz_cols <- c("X", "Y", "Z")
  mat_x <- x@points[, xyz_cols, drop = FALSE]
  mat_y <- y@points[, xyz_cols, drop = FALSE]
  hausdorff_distance_cpp(mat_x, mat_y)
}

# --- method: bundle ---------------------------------------------------------

#' [compute_hausdorff_distance()] method for `bundle` objects
#'
#' Dispatches to one of three behaviours depending on `y`:
#'
#' - `y = NULL`: pairwise distances within `x` as a [`dist`][stats::dist]
#'   object of size \eqn{n}, computed in C++ via a single linear loop.
#' - `y` is a [streamline]: numeric vector of distances from `y` to each
#'   streamline in `x`.
#' - `y` is a [bundle]: [`dist`][stats::dist] object for the concatenation
#'   of all streamlines from `x` and `y`.
#'
#' @param x A [bundle] object.
#' @param y `NULL`, a [streamline], or a [bundle].
#' @returns
#'   - A [`dist`][stats::dist] object of size `x@n_streamlines` when `y` is
#'     `NULL` or a [bundle]. The lower triangle stores all pairwise symmetric
#'     Hausdorff distances (computed in C++). Use [as.matrix()] to obtain the
#'     full \eqn{n \times n} matrix.
#'   - A numeric vector of length `x@n_streamlines` when `y` is a [streamline].
#' @seealso [compute_hausdorff_distance()]
#' @name compute_hausdorff_distance-fiber-bundle-method
#' @aliases compute_hausdorff_distance,fiber::bundle-method
#' @usage NULL
#' @examples
#' pts1 <- matrix(runif(30), ncol = 3)
#' colnames(pts1) <- c("X", "Y", "Z")
#' pts2 <- matrix(runif(30), ncol = 3)
#' colnames(pts2) <- c("X", "Y", "Z")
#' sl1 <- streamline(points = pts1)
#' sl2 <- streamline(points = pts2)
#' b <- bundle(streamlines = list(sl1, sl2))
#'
#' # pairwise dist object (size 2)
#' compute_hausdorff_distance(b)
#' as.matrix(compute_hausdorff_distance(b))
#'
#' # distances from sl1 to each streamline in b
#' compute_hausdorff_distance(b, sl1)
S7::method(compute_hausdorff_distance, bundle) <- function(x, y = NULL) {
  xyz_cols <- c("X", "Y", "Z")
  if (is.null(y)) {
    # --- bundle x missing: dist object via C++ single-loop ------------------
    sls <- x@streamlines
    n <- length(sls)
    mats <- lapply(sls, function(sl) sl@points[, xyz_cols, drop = FALSE])
    dv <- pairwise_hausdorff_cpp(mats)
    structure(
      dv,
      class = "dist",
      Size = n,
      Diag = FALSE,
      Upper = FALSE,
      method = "hausdorff"
    )
  } else if (is_streamline(y)) {
    # --- bundle x streamline: vector of distances ---------------------------
    mat_y <- y@points[, xyz_cols, drop = FALSE]
    vapply(
      x@streamlines,
      function(sl) {
        hausdorff_distance_cpp(
          sl@points[, xyz_cols, drop = FALSE],
          mat_y
        )
      },
      numeric(1L)
    )
  } else if (is_bundle(y)) {
    # --- bundle x bundle: combined dist object ------------------------------
    combined <- bind_bundles(x, y)
    compute_hausdorff_distance(combined)
  } else {
    cli::cli_abort(c(
      "{.arg y} must be {.code NULL}, a {.cls fiber::streamline}, or a
      {.cls fiber::bundle} when {.arg x} is a {.cls fiber::bundle}.",
      "x" = "Got {.obj_type_friendly {y}} instead."
    ))
  }
}

# --- catch-all method: unsupported x types ----------------------------------

#' [compute_hausdorff_distance()] catch-all method
#'
#' @usage NULL
#' @returns Does not return a value; always throws an error for unsupported
#'   input types.
S7::method(compute_hausdorff_distance, S7::class_any) <- function(x, y = NULL) {
  cli::cli_abort(c(
    "{.arg x} must be a {.cls fiber::streamline} or {.cls fiber::bundle}
    object.",
    "x" = "Got {.obj_type_friendly {x}} instead.",
    "i" = "Supported combinations are: {.cls streamline} + {.cls streamline},
    {.cls bundle} (no {.arg y}), {.cls bundle} + {.cls streamline}, and
    {.cls bundle} + {.cls bundle}."
  ))
}

library(fiber)

set.seed(42)
pts1 <- matrix(runif(30), ncol = 3, dimnames = list(NULL, c("X", "Y", "Z")))
pts2 <- matrix(runif(30), ncol = 3, dimnames = list(NULL, c("X", "Y", "Z")))
sl1 <- streamline(pts1)
sl2 <- streamline(pts2)
b <- bundle(list(sl1, sl2))

not_sl <- list(1, 2, 3)

# ---- get_euclidean_length ---------------------------------------------------

el <- get_euclidean_length(sl1)
expect_true(is.numeric(el) && length(el) == 1L && el >= 0)
expect_error(get_euclidean_length(not_sl))

# ---- get_curvilinear_length -------------------------------------------------

cl <- get_curvilinear_length(sl1)
expect_true(is.numeric(cl) && length(cl) == 1L && cl >= 0)
expect_error(get_curvilinear_length(not_sl))

# curvilinear >= euclidean
expect_true(cl >= el)

# ---- get_sinuosity ----------------------------------------------------------

sin_val <- get_sinuosity(sl1)
expect_true(is.numeric(sin_val) && sin_val >= 1)
expect_error(get_sinuosity(not_sl))

# ---- get_curvature ----------------------------------------------------------

curv <- get_curvature(sl1)
expect_equal(length(curv), sl1@n_points)
expect_true(all(curv >= 0))
expect_error(get_curvature(not_sl))

# ---- get_torsion ------------------------------------------------------------

tors <- get_torsion(sl1)
expect_equal(length(tors), sl1@n_points)
expect_true(is.numeric(tors))
expect_error(get_torsion(not_sl))

# ---- add_shape_descriptors: streamline --------------------------------------

sl_desc <- add_shape_descriptors(
  sl1,
  descriptors = c("euclidean_length", "curvilinear_length", "sinuosity")
)
expect_true(is_streamline(sl_desc))
expect_true(!is.null(sl_desc@streamline_data$euclidean_length))
expect_true(!is.null(sl_desc@streamline_data$curvilinear_length))
expect_true(!is.null(sl_desc@streamline_data$sinuosity))

# curvature and torsion go into point_data
sl_desc2 <- add_shape_descriptors(sl1, descriptors = c("curvature", "torsion"))
expect_equal(length(sl_desc2@point_data$curvature), sl1@n_points)
expect_equal(length(sl_desc2@point_data$torsion), sl1@n_points)

# unknown descriptor emits a warning
expect_warning(add_shape_descriptors(sl1, descriptors = "unknown_desc"))

# ---- add_shape_descriptors: bundle ------------------------------------------

b_desc <- add_shape_descriptors(
  b,
  descriptors = c("euclidean_length", "curvilinear_length", "sinuosity")
)
expect_true(is_bundle(b_desc))
expect_true(!is.null(b_desc[[1L]]@streamline_data$euclidean_length))
expect_true(!is.null(b_desc[[2L]]@streamline_data$euclidean_length))

# ---- compute_hausdorff_distance: streamline x streamline --------------------

d_sl <- compute_hausdorff_distance(sl1, sl2)
expect_true(is.numeric(d_sl) && length(d_sl) == 1L && d_sl >= 0)

# y not a streamline when x is streamline -> error
expect_error(compute_hausdorff_distance(sl1, not_sl))
expect_error(compute_hausdorff_distance(sl1))

# ---- compute_hausdorff_distance: bundle x missing ---------------------------

d_b <- compute_hausdorff_distance(b)
expect_true(inherits(d_b, "dist"))
expect_equal(attr(d_b, "Size"), 2L)

# as.matrix gives 2x2
dm <- as.matrix(d_b)
expect_equal(dim(dm), c(2L, 2L))
expect_equal(dm[1L, 1L], 0)

# ---- compute_hausdorff_distance: bundle x streamline ------------------------

d_bsl <- compute_hausdorff_distance(b, sl1)
expect_equal(length(d_bsl), 2L)
expect_true(all(d_bsl >= 0))

# ---- compute_hausdorff_distance: bundle x bundle ----------------------------

b2 <- bundle(list(sl2))
d_bb <- compute_hausdorff_distance(b, b2)
expect_true(inherits(d_bb, "dist"))
expect_equal(attr(d_bb, "Size"), 3L)

# ---- compute_hausdorff_distance: invalid y type ----------------------------

expect_error(compute_hausdorff_distance(b, not_sl))

# ---- compute_hausdorff_distance: catch-all (non-streamline/bundle x) -------

expect_error(compute_hausdorff_distance(42))

#' @include bundle.R streamline.R
NULL

# ---- reparametrize ----------------------------------------------------------

#' Reparametrize a streamline or bundle onto a uniform arc-length grid
#'
#' @description
#' `reparametrize()` is an S7 generic that resamples the 3-D coordinates (and
#' any `@point_data` attributes) of a tractography object onto a uniform
#' arc-length grid using linear interpolation, with methods available for the
#' following classes:
#'
#' `r doclisting::methods_list("reparametrize")`
#'
#' @param x A [streamline] or [bundle] object.
#' @param n_points Number of equally-spaced arc-length points to use.
#'
#'   - For a single [streamline], defaults to `nrow(x@points)`.
#'   - For a [bundle], defaults to the rounded mean number of points across
#'     all streamlines.
#'
#'   Pass `NULL` to use these defaults explicitly.
#' @returns An object of the same class as `x` reparametrized onto the new
#'   grid.
#' @export
#' @examples
#' # reparametrize a single streamline to 10 points
#' pts <- matrix(runif(30), ncol = 3)
#' colnames(pts) <- c("X", "Y", "Z")
#' sl <- streamline(points = pts)
#' sl_reparam <- reparametrize(sl, n_points = 10)
#' # reparametrize a bundle to the mean number of points across its streamlines
#' sl1 <- streamline(points = pts)
#' pts2 <- matrix(runif(60), ncol = 3)
#' colnames(pts2) <- c("X", "Y", "Z")
#' sl2 <- streamline(points = pts2)
#' b <- bundle(streamlines = list(sl1, sl2))
#' bundle_reparam <- reparametrize(b)
reparametrize <- S7::new_generic(
  "reparametrize",
  "x",
  function(x, n_points = NULL) {
    S7::S7_dispatch()
  }
)

# ---- method: streamline -----------------------------------------------------

#' [reparametrize()] method for `streamline` objects
#'
#' Resamples a single [streamline] onto a uniform arc-length grid. See
#' [reparametrize()] for the full parameter documentation.
#'
#' @param x A [streamline] object.
#' @inheritParams reparametrize
#' @returns A [streamline] reparametrized onto the new grid. The returned
#'   object has the same class as the input but with `@points` resampled to
#'   exactly `n_points` rows and all `@point_data` vectors resampled
#'   correspondingly via linear interpolation.
#' @seealso [reparametrize()]
#' @name reparametrize-fiber-streamline-method
#' @aliases reparametrize,fiber::streamline-method
#' @usage NULL
#' @examples
#' pts <- matrix(runif(30), ncol = 3)
#' colnames(pts) <- c("X", "Y", "Z")
#' sl <- streamline(points = pts, point_data = list(FA = runif(10)))
#' sl_reparam <- reparametrize(sl, n_points = 20)
#' sl_reparam@n_points  # 20
S7::method(reparametrize, streamline) <- function(x, n_points = NULL) {
  pts <- x@points
  s <- .arc_length(pts)
  n <- if (is.null(n_points)) nrow(pts) else as.integer(n_points)

  if (n < 2L) {
    cli::cli_abort("{.arg n_points} must be at least 2.")
  }

  s_new <- seq(s[1L], s[length(s)], length.out = n)

  new_pts <- cbind(
    X = .approx1(s, pts[, "X"], s_new),
    Y = .approx1(s, pts[, "Y"], s_new),
    Z = .approx1(s, pts[, "Z"], s_new)
  )
  # preserve any extra columns beyond X, Y, Z
  extra_cols <- setdiff(colnames(pts), c("X", "Y", "Z"))
  if (length(extra_cols) > 0L) {
    extra <- vapply(
      extra_cols,
      function(col) {
        .approx1(s, pts[, col], s_new)
      },
      numeric(n)
    )
    new_pts <- cbind(new_pts, extra)
  }

  # Only numeric point_data can be interpolated; non-numeric entries are
  # dropped with a warning because they have no natural interpolant.
  new_pd <- list()
  for (nm in names(x@point_data)) {
    v <- x@point_data[[nm]]
    if (is.numeric(v)) {
      new_pd[[nm]] <- .approx1(s, v, s_new)
    } else {
      cli::cli_warn(c(
        "!" = "Non-numeric {.field @point_data} attribute {.val {nm}} \\
               cannot be interpolated and will be dropped.",
        "i" = "Only numeric per-point attributes are preserved by \\
               {.fn reparametrize}."
      ))
    }
  }

  streamline(
    points = new_pts,
    point_data = new_pd,
    streamline_data = x@streamline_data
  )
}

# ---- method: bundle ---------------------------------------------------------

#' [reparametrize()] method for `bundle` objects
#'
#' Resamples every [streamline] inside a [bundle] onto a common uniform
#' arc-length grid. See [reparametrize()] for the full parameter documentation.
#'
#' @param x A [bundle] object.
#' @inheritParams reparametrize
#' @returns A [bundle] reparametrized onto the new grid. Every streamline in the
#'   returned bundle has exactly `n_points` rows in `@points` (defaulting to
#'   the rounded mean number of points across all streamlines when `n_points`
#'   is `NULL`).
#' @seealso [reparametrize()]
#' @name reparametrize-fiber-bundle-method
#' @aliases reparametrize,fiber::bundle-method
#' @usage NULL
#' @examples
#' pts1 <- matrix(runif(30), ncol = 3)
#' colnames(pts1) <- c("X", "Y", "Z")
#' pts2 <- matrix(runif(60), ncol = 3)
#' colnames(pts2) <- c("X", "Y", "Z")
#' b <- bundle(streamlines = list(streamline(points = pts1),
#'                                 streamline(points = pts2)))
#' b_reparam <- reparametrize(b, n_points = 15)
#' b_reparam[[1]]@n_points  # 15
#' b_reparam[[2]]@n_points  # 15
S7::method(reparametrize, bundle) <- function(x, n_points = NULL) {
  if (length(x@streamlines) == 0L) {
    return(x)
  }

  n <- if (is.null(n_points)) {
    round(mean(vapply(
      x@streamlines,
      function(sl) nrow(sl@points),
      integer(1L)
    )))
  } else {
    as.integer(n_points)
  }

  new_sls <- lapply(x@streamlines, reparametrize, n_points = n)
  bundle(streamlines = new_sls, bundle_data = x@bundle_data)
}

# ---- bind_bundles ------------------------------------------------------------

#' Combine streamlines and/or bundles into a single bundle
#'
#' Accepts any mix of [streamline] and [bundle] objects. All streamlines are
#' collected into a flat list and wrapped in a new [bundle]. `bundle_data`
#' from the first [bundle] argument (if any) is preserved; pass your own via
#' the `bundle_data` argument to override.
#'
#' @param ... One or more [streamline] or [bundle] objects.
#' @param bundle_data A named list of bundle-level metadata to attach to the
#'   resulting [bundle]. Defaults to an empty list (or the `bundle_data` of
#'   the first [bundle] input if one is present and `bundle_data` is not
#'   supplied).
#' @returns A [bundle] containing all input streamlines.
#' @export
#' @examples
#' pts <- matrix(runif(15), ncol = 3, dimnames = list(NULL, c("X", "Y", "Z")))
#' sl1 <- streamline(points = pts)
#' sl2 <- streamline(points = pts)
#' b1 <- bundle(streamlines = list(sl1))
#' b2 <- bundle(streamlines = list(sl2))
#'
#' # combine two bundles
#' b_all <- bind_bundles(b1, b2)
#' b_all@n_streamlines  # 2
#'
#' # mix a bundle and a loose streamline
#' b_mixed <- bind_bundles(b1, sl2)
#' b_mixed@n_streamlines  # 2
bind_bundles <- function(..., bundle_data = NULL) {
  inputs <- list(...)
  if (length(inputs) == 0L) {
    cli::cli_abort("At least one argument is required.")
  }

  sls <- list()
  first_bd <- list()
  found_bundle <- FALSE

  for (obj in inputs) {
    if (is_streamline(obj)) {
      sls <- c(sls, list(obj))
    } else if (is_bundle(obj)) {
      sls <- c(sls, obj@streamlines)
      if (!found_bundle) {
        first_bd <- obj@bundle_data
        found_bundle <- TRUE
      }
    } else {
      cli::cli_abort(
        "Each argument must be a {.cls fiber::streamline} or {.cls fiber::bundle}."
      )
    }
  }

  bd <- if (!is.null(bundle_data)) bundle_data else first_bd
  bundle(streamlines = sls, bundle_data = bd)
}

# ---- as_bundle_set ----------------------------------------------------------

#' Coerce an object to a bundle_set
#'
#' `as_bundle_set()` converts a supported object into a [bundle_set].
#'
#' Currently supported input classes:
#' - [bundle_set]: returned unchanged.
#' - [bundle]: wrapped in a single-element [bundle_set]. An optional `name`
#'   argument sets the element name (defaults to `"bundle_1"`).
#'
#' @param x An object to coerce.
#' @param ... Additional arguments passed to methods (e.g. `name` for bundles).
#' @returns A [bundle_set] object.
#' @seealso [bundle_set()], [bind_bundle_sets()]
#' @export
#' @examples
#' pts <- matrix(runif(15), ncol = 3, dimnames = list(NULL, c("X", "Y", "Z")))
#' b <- bundle(streamlines = list(streamline(points = pts)))
#' bs <- as_bundle_set(b, name = "sub-01")
#' bs@n_bundles  # 1
as_bundle_set <- S7::new_generic("as_bundle_set", "x")

S7::method(as_bundle_set, bundle_set) <- function(x, ...) x

S7::method(as_bundle_set, bundle) <- function(x, ..., name = "bundle_1") {
  nms <- list(x)
  names(nms) <- name
  bundle_set(bundles = nms)
}

S7::method(as_bundle_set, S7::class_any) <- function(x, ...) {
  cli::cli_abort(
    "Don't know how to coerce an object of class {.cls {class(x)[1L]}} \\
     to a {.cls bundle_set}."
  )
}

# ---- bind_bundle_sets -------------------------------------------------------

#' Combine bundles and/or bundle_sets into a single bundle_set
#'
#' Accepts any mix of named [bundle] objects (passed as `name = bundle`) or
#' [bundle_set] objects. All bundles are collected into a flat named list and
#' wrapped in a new [bundle_set].
#'
#' @param ... Named [bundle] objects or [bundle_set] objects to combine. Each
#'   bare [bundle] argument must be **named** so that its label in the
#'   resulting set is unambiguous.
#' @param set_data A named list of set-level metadata to attach to the
#'   resulting [bundle_set]. Defaults to the `set_data` of the first
#'   [bundle_set] input (if present) or an empty list.
#' @returns A [bundle_set] containing all input bundles.
#' @export
#' @examples
#' pts <- matrix(runif(15), ncol = 3, dimnames = list(NULL, c("X", "Y", "Z")))
#' b1 <- bundle(streamlines = list(streamline(points = pts)))
#' b2 <- bundle(streamlines = list(streamline(points = pts)))
#'
#' # two named bare bundles
#' bs <- bind_bundle_sets("sub-01" = b1, "sub-02" = b2)
#' bs@n_bundles   # 2
#' bs@bundle_names  # c("sub-01", "sub-02")
#'
#' # combine two bundle_sets
#' bs1 <- bundle_set(list("sub-01" = b1))
#' bs2 <- bundle_set(list("sub-02" = b2))
#' bs_all <- bind_bundle_sets(bs1, bs2)
#' bs_all@n_bundles  # 2
bind_bundle_sets <- function(..., set_data = NULL) {
  inputs <- list(...)
  if (length(inputs) == 0L) {
    cli::cli_abort("At least one argument is required.")
  }

  nms <- names(inputs)
  all_bundles <- list()
  first_sd <- list()
  found_set <- FALSE

  for (i in seq_along(inputs)) {
    obj <- inputs[[i]]
    nm <- if (!is.null(nms)) nms[[i]] else ""
    if (is_bundle(obj)) {
      if (is.null(nm) || nm == "") {
        cli::cli_abort(c(
          "Bare {.cls fiber::bundle} arguments must be named.",
          "i" = "Use {.code bind_bundle_sets(\"sub-01\" = b1, \"sub-02\" = b2)}."
        ))
      }
      all_bundles[[nm]] <- obj
    } else if (is_bundle_set(obj)) {
      new_bds <- obj@bundles
      if (length(new_bds) > 0L) {
        dups <- intersect(names(all_bundles), names(new_bds))
        if (length(dups) > 0L) {
          cli::cli_abort(c(
            "Duplicate bundle name{?s}: {.val {dups}}.",
            "i" = "Each bundle in a {.cls bundle_set} must have a unique name."
          ))
        }
        all_bundles <- c(all_bundles, new_bds)
      }
      if (!found_set) {
        first_sd <- obj@set_data
        found_set <- TRUE
      }
    } else {
      cli::cli_abort(
        "Each argument must be a named {.cls fiber::bundle} or a \\
         {.cls fiber::bundle_set}."
      )
    }
  }

  sd <- if (!is.null(set_data)) set_data else first_sd
  bundle_set(bundles = all_bundles, set_data = sd)
}

# ---- S7 class: streamline ---------------------------------------------------

#' Streamline S7 class
#'
#' @description
#' A `streamline` represents a single fibre tract. It stores three data
#' compartments that mirror the conceptual levels found in tractography file
#' formats:
#'
#' - `@points` — an \eqn{n \times 3} numeric matrix whose columns are named
#'   `"X"`, `"Y"`, and `"Z"`, holding the ordered 3-D coordinates of the
#'   \eqn{n} points along the tract.
#' - `@point_data` — a named list of numeric vectors, each of length \eqn{n},
#'   holding per-point scalar attributes (e.g. fractional anisotropy sampled
#'   at every point).
#' - `@streamline_data` — a named list of numeric scalars (length-1 vectors)
#'   holding per-streamline attributes (e.g. a tract-level weight or mean FA).
#'
#' @param points A numeric matrix with columns `"X"`, `"Y"`, and `"Z"`.
#' @param point_data A named list of per-point numeric vectors.
#' @param streamline_data A named list of per-streamline numeric scalars.
#' @prop n_points An integer scalar giving the number of points in the streamline (read-only).
#' @prop point_attributes A character vector of the names of the per-point attributes (read-only).
#' @prop streamline_attributes A character vector of the names of the per-streamline attributes (read-only).
#'
#' @returns A `streamline` S7 object.
#' @section Methods for standard generics:
#' The following methods are defined for `streamline` objects:
#' - `format(x, ...)`: Returns a compact character string such as
#'   `<streamline [10 pts] | point: FA>`.
#' - `print(x, ...)`: Prints the formatted string to the console and
#'   invisibly returns `x`.
#' @export
#' @examples
#' # Create a streamline with 5 points and some attributes
#' sl <- streamline(
#'   points = matrix(
#'     c(0, 0, 0,
#'       1, 0, 0,
#'       1, 1, 0,
#'       1, 1, 1,
#'       0, 1, 1),
#'     ncol = 3,
#'     byrow = TRUE,
#'     dimnames = list(NULL, c("X", "Y", "Z"))
#'   ),
#'   point_data = list(FA = c(0.5, 0.6, 0.7, 0.8, 0.9)),
#'   streamline_data = list(mean_FA = 0.7)
#' )
#' sl@n_points  # 5
#' sl@point_attributes  # "FA"
#' sl@streamline_attributes  # "mean_FA"
#'
#' # format() and print() methods
#' format(sl)  # "<streamline [5 pts] | point: FA | streamline: mean_FA>"
#' print(sl)
streamline <- S7::new_class(
  name = "streamline",
  package = "fiber",
  properties = list(
    points = S7::class_any,
    point_data = S7::class_list,
    streamline_data = S7::class_list,
    n_points = S7::new_property(
      class = S7::class_integer,
      getter = function(self) nrow(self@points)
    ),
    point_attributes = S7::new_property(
      class = S7::class_character,
      getter = function(self) names(self@point_data)
    ),
    streamline_attributes = S7::new_property(
      class = S7::class_character,
      getter = function(self) names(self@streamline_data)
    )
  ),
  validator = function(self) {
    if (!is.matrix(self@points) || !is.numeric(self@points)) {
      return("@points must be a numeric matrix.")
    }
    cn <- colnames(self@points)
    if (is.null(cn) || !all(c("X", "Y", "Z") %in% cn)) {
      return("@points must have column names including 'X', 'Y', and 'Z'.")
    }
    n_pts <- self@n_points
    for (nm in self@point_attributes) {
      v <- self@point_data[[nm]]
      if (length(v) != n_pts) {
        return(cli::format_inline(
          "@point_data[[\"{nm}\"]] must be a vector of length {n_pts}."
        ))
      }
    }
    for (nm in self@streamline_attributes) {
      v <- self@streamline_data[[nm]]
      if (length(v) != 1L) {
        return(cli::format_inline(
          "@streamline_data[[\"{nm}\"]] must be a scalar (length 1)."
        ))
      }
    }
    NULL
  }
)

# ---- predicate --------------------------------------------------------------

#' Test whether an object is a streamline
#'
#' @param x An object.
#' @returns `TRUE` if `x` is of class [streamline], otherwise `FALSE`.
#' @export
#' @examples
#' pts <- matrix(runif(15), ncol = 3, dimnames = list(NULL, c("X", "Y", "Z")))
#' sl <- streamline(points = pts)
#' is_streamline(sl)     # TRUE
#' is_streamline(42)     # FALSE
is_streamline <- function(x) S7::S7_inherits(x, streamline)

# ---- helper: number of points -----------------------------------------------

#' Number of points in a streamline
#'
#' @param x A [streamline] object.
#' @returns An integer scalar giving the number of points.
#' @keywords internal
n_points <- function(x) x@n_points

# ---- format / print ---------------------------------------------------------

S7::method(format, streamline) <- function(x, ...) {
  pd <- x@point_attributes
  sld <- x@streamline_attributes
  pd_str <- if (length(pd) > 0L) {
    paste0(" | point: ", paste(pd, collapse = ", "))
  } else {
    ""
  }
  sld_str <- if (length(sld) > 0L) {
    paste0(" | streamline: ", paste(sld, collapse = ", "))
  } else {
    ""
  }
  paste0("<streamline [", x@n_points, " pts]", pd_str, sld_str, ">")
}

S7::method(print, streamline) <- function(x, ...) {
  cat(format(x, ...), "\n")
  invisible(x)
}

library(fiber)

# ---- streamline ---------------------------------------------------------

# valid matrix
pts <- matrix(
  c(1, 2, 3, 4, 5, 6, 7, 8, 9),
  ncol = 3,
  dimnames = list(NULL, c("X", "Y", "Z"))
)
sl <- streamline(pts)
expect_true(is_streamline(sl))
expect_true(is.matrix(sl@points))

# not a matrix -> error
expect_error(streamline("not a matrix"))

# no colnames -> error
expect_error(streamline(matrix(1:9, ncol = 3)))

# missing required column names -> error
bad <- matrix(1:9, ncol = 3, dimnames = list(NULL, c("A", "B", "C")))
expect_error(streamline(bad))

# point_data length mismatch -> error
expect_error(streamline(pts, point_data = list(FA = c(0.1, 0.2))))

# streamline_data non-scalar -> error
expect_error(streamline(pts, streamline_data = list(weight = c(1, 2))))

# ---- with point_data and streamline_data ------------------------------------

sl_pd <- streamline(pts, point_data = list(FA = c(0.1, 0.2, 0.3)))
sl_sld <- streamline(pts, streamline_data = list(weight = 0.5))
sl_all <- streamline(
  pts,
  point_data = list(FA = c(0.1, 0.2, 0.3)),
  streamline_data = list(weight = 0.5)
)

expect_true(is_streamline(sl_pd))
expect_equal(sl_pd@point_data[["FA"]], c(0.1, 0.2, 0.3))
expect_equal(sl_sld@streamline_data[["weight"]], 0.5)
expect_equal(names(sl_all@point_data), "FA")
expect_equal(names(sl_all@streamline_data), "weight")

# ---- is_streamline ----------------------------------------------------------

expect_true(is_streamline(sl))
expect_false(is_streamline(pts))
expect_false(is_streamline(list()))

# ---- format.streamline ------------------------------------------------------

f <- format(sl)
expect_true(grepl("streamline", f))
expect_true(grepl("3 pts", f))

f_pd <- format(sl_pd)
expect_true(grepl("point: FA", f_pd))

f_all <- format(sl_all)
expect_true(grepl("point: FA", f_all))
expect_true(grepl("streamline: weight", f_all))

# ---- print.streamline -------------------------------------------------------

expect_stdout(print(sl), "streamline")

library(fiber)

pts <- matrix(
  c(1, 2, 3, 4, 5, 6, 7, 8, 9),
  ncol = 3,
  dimnames = list(NULL, c("X", "Y", "Z"))
)
sl <- streamline(pts)

# ---- reparametrize: streamline ----------------------------------------------

# default n_points (same as input)
sl_r <- reparametrize(sl)
expect_true(is_streamline(sl_r))
expect_equal(sl_r@n_points, sl@n_points)

# explicit n_points
sl_r10 <- reparametrize(sl, n_points = 10L)
expect_equal(sl_r10@n_points, 10L)

# n_points < 2 -> error
expect_error(reparametrize(sl, n_points = 1L))

# preserves point_data via interpolation
sl_pd <- streamline(pts, point_data = list(FA = c(0.1, 0.2, 0.3)))
sl_pd_r <- reparametrize(sl_pd, n_points = 5L)
expect_equal(length(sl_pd_r@point_data$FA), 5L)

# extra columns in points matrix are preserved
pts_extra <- cbind(pts, W = c(1.0, 2.0, 3.0))
colnames(pts_extra) <- c("X", "Y", "Z", "W")
sl_extra <- streamline(pts_extra)
sl_extra_r <- reparametrize(sl_extra, n_points = 5L)
expect_true("W" %in% colnames(sl_extra_r@points))

# ---- reparametrize: bundle --------------------------------------------------

pts2 <- matrix(runif(30), ncol = 3, dimnames = list(NULL, c("X", "Y", "Z")))
sl2 <- streamline(pts2)
b <- bundle(list(sl, sl2))

# default n_points (rounded mean)
b_r <- reparametrize(b)
expect_true(is_bundle(b_r))
expect_equal(length(b_r@streamlines), 2L)

# explicit n_points
b_r5 <- reparametrize(b, n_points = 5L)
expect_equal(b_r5[[1L]]@n_points, 5L)
expect_equal(b_r5[[2L]]@n_points, 5L)

# empty bundle returns unchanged
b0 <- bundle(list())
b0_r <- reparametrize(b0)
expect_equal(b0_r@n_streamlines, 0L)

# ---- bind_bundles ------------------------------------------------------------

sl1 <- streamline(pts)
sl3 <- streamline(pts2)
b1 <- bundle(list(sl1), bundle_data = list(origin = "A"))
b2 <- bundle(list(sl3))

# two bundles: preserves first bundle's bundle_data
b_all <- bind_bundles(b1, b2)
expect_true(is_bundle(b_all))
expect_equal(b_all@n_streamlines, 2L)
expect_equal(b_all@bundle_data[["origin"]], "A")

# mix bundle and streamline
b_mixed <- bind_bundles(b1, sl3)
expect_equal(b_mixed@n_streamlines, 2L)

# explicit bundle_data overrides
b_override <- bind_bundles(b1, b2, bundle_data = list(origin = "override"))
expect_equal(b_override@bundle_data[["origin"]], "override")

# no arguments -> error
expect_error(bind_bundles())

# invalid argument -> error
expect_error(bind_bundles(b1, "not_valid"))

library(fiber)

# ---- bundle constructor -----------------------------------------------------

# valid matrix
pts <- matrix(
  c(1, 2, 3, 4, 5, 6, 7, 8, 9),
  ncol = 3,
  dimnames = list(NULL, c("X", "Y", "Z"))
)
sl <- streamline(pts)
sl_pd <- streamline(pts, point_data = list(FA = c(0.1, 0.2, 0.3)))
sl_all <- streamline(
  pts,
  point_data = list(FA = c(0.1, 0.2, 0.3)),
  streamline_data = list(weight = 0.5)
)

b <- bundle(list(sl, sl_pd))
expect_true(is_bundle(b))

# non-list -> error
expect_error(bundle(sl))

# list with non-streamline element -> error
expect_error(bundle(list(sl, "not a streamline")))

# empty list is allowed (zero-streamline bundle)
b0 <- bundle(list())
expect_true(is_bundle(b0))

# bundle_data is stored
b_meta <- bundle(list(sl), bundle_data = list(origin = "phantom"))
expect_equal(b_meta@bundle_data[["origin"]], "phantom")

# ---- is_bundle --------------------------------------------------------------

expect_true(is_bundle(b))
expect_false(is_bundle(sl))
expect_false(is_bundle(list(sl)))

# ---- format.bundle ----------------------------------------------------------

f0 <- format(b0)
expect_true(grepl("0 streamlines", f0))

b_plain <- bundle(list(sl, sl))
fp <- format(b_plain)
expect_true(grepl("streamlines", fp))

b_pd <- bundle(list(sl_pd, sl_pd))
fp_pd <- format(b_pd)
expect_true(grepl("point: FA", fp_pd))

b_all <- bundle(list(sl_all, sl_all))
fp_all <- format(b_all)
expect_true(grepl("point: FA", fp_all))
expect_true(grepl("streamline: weight", fp_all))

# ---- print.bundle -----------------------------------------------------------

expect_stdout(print(b_plain), "streamlines")

# ---- length.bundle ----------------------------------------------------------

expect_equal(length(b), 2L)
expect_equal(length(b0), 0L)

# ---- indexing ---------------------------------------------------------------

expect_true(is_streamline(b@streamlines[[1L]]))

# [[ extracts a streamline via S7 method
expect_true(is_streamline(b[[1L]]))

# [ returns a bundle (subset) preserving bundle_data
b_sub <- b_meta[1L]
expect_true(is_bundle(b_sub))
expect_equal(b_sub@bundle_data[["origin"]], "phantom")
expect_equal(b_sub@n_streamlines, 1L)

# ---- bundle_attributes getter -----------------------------------------------

expect_null(b@bundle_attributes)
b_attr <- bundle(list(sl), bundle_data = list(a = 1, b = 2))
expect_equal(b_attr@bundle_attributes, c("a", "b"))

# ---- format with bundle_data ------------------------------------------------

b_bd <- bundle(list(sl), bundle_data = list(origin = "test"))
fb_bd <- format(b_bd)
expect_true(grepl("bundle: origin", fb_bd))

library(fiber)

# ---- helpers ----------------------------------------------------------------

make_sl <- function(n = 5L) {
  pts <- matrix(as.numeric(seq_len(n * 3L)), nrow = n, ncol = 3L)
  colnames(pts) <- c("X", "Y", "Z")
  streamline(points = pts)
}

# ---- as_streamline ----------------------------------------------------------

sl <- make_sl()

# identity: as_streamline(streamline) returns the same object
expect_identical(as_streamline(sl), sl)

# as_streamline applied to a single-streamline bundle
b1 <- bundle(streamlines = list(sl))
result_sl <- as_streamline(b1)
expect_true(is_streamline(result_sl))
expect_identical(result_sl@points, sl@points)

# as_streamline on a multi-streamline bundle should error
b2 <- bundle(streamlines = list(sl, make_sl(4L)))
expect_error(as_streamline(b2))

# as_streamline on an unknown class should error
expect_error(as_streamline(42L))
expect_error(as_streamline("hello"))

# ---- as_bundle --------------------------------------------------------------

# wrapping: as_bundle(streamline) gives a 1-streamline bundle
b_from_sl <- as_bundle(sl)
expect_true(is_bundle(b_from_sl))
expect_equal(b_from_sl@n_streamlines, 1L)
expect_identical(b_from_sl[[1L]]@points, sl@points)

# identity: as_bundle(bundle) returns the same object
b_orig <- bundle(streamlines = list(sl))
expect_identical(as_bundle(b_orig), b_orig)

# as_bundle on an unknown class should error
expect_error(as_bundle(TRUE))

# ---- as_dwifiber (only when dti is available) -------------------------------

if (requireNamespace("dti", quietly = TRUE)) {
  rgl_use_null <- Sys.getenv("RGL_USE_NULL", unset = NA)
  Sys.setenv(RGL_USE_NULL = "true") # avoid rgl warnings

  # bundle -> dwiFiber and back
  sl1 <- make_sl(6L)
  sl2 <- make_sl(4L)
  b <- bundle(
    streamlines = list(sl1, sl2),
    bundle_data = list(
      method = "LINEPROP",
      minfa = 0.3,
      maxangle = 30,
      voxelext = c(2, 2, 2)
    )
  )

  dfi <- as_dwifiber(b)
  expect_true(is(dfi, "dwiFiber"))
  expect_equal(length(dfi@startind), 2L)
  expect_equal(dfi@method, "LINEPROP")
  expect_equal(dfi@minfa, 0.3)
  expect_equal(dfi@maxangle, 30)
  expect_equal(dfi@voxelext, c(2, 2, 2))

  # Round-trip: dwiFiber -> bundle -> dwiFiber
  b_rt <- as_bundle(dfi)
  expect_true(is_bundle(b_rt))
  expect_equal(b_rt@n_streamlines, 2L)

  # Points are recovered correctly
  expect_identical(
    b_rt[[1L]]@points,
    sl1@points
  )

  # as_dwifiber(streamline) same as wrapping in a bundle first
  dfi_sl <- as_dwifiber(sl1)
  expect_true(is(dfi_sl, "dwiFiber"))
  expect_equal(length(dfi_sl@startind), 1L)

  Sys.setenv(RGL_USE_NULL = rgl_use_null) # restore previous state
}

# ---- S7 class: bundle -------------------------------------------------------

#' Bundle S7 class
#'
#' @description
#' A `bundle` is an ordered collection of [streamline] objects representing a
#' tractogram or white-matter bundle. It stores two compartments:
#'
#' - `@streamlines` — a list of [streamline] objects.
#' - `@bundle_data` — a named list of bundle-level metadata (arbitrary R
#'   objects, e.g. the affine transform used during tracking).
#'
#' @param streamlines A list of [streamline] objects.
#' @param bundle_data A named list of bundle-level metadata.
#' @prop n_streamlines An integer scalar giving the number of streamlines in the bundle (read-only).
#' @prop bundle_attributes A character vector of the names of the bundle-level attributes (read-only).
#'
#' @returns A `bundle` S7 object.
#' @section Methods for standard generics:
#' The following methods are defined for `bundle` objects:
#' - `format(x, ...)`: Returns a compact character string such as
#'   `<bundle [2 streamlines | 10–20 pts/streamline]>`.
#' - `print(x, ...)`: Prints the formatted string to the console and
#'   invisibly returns `x`.
#' - `length(x)`: Returns the number of streamlines (equivalent to
#'   `x@n_streamlines`).
#' - `x[[i]]`: Extracts the `i`-th [streamline] from the bundle.
#' - `x[i]`: Returns a new [bundle] containing only the selected
#'   streamlines, preserving `@bundle_data`.
#' @export
#' @examples
#' pts <- matrix(runif(15), ncol = 3, dimnames = list(NULL, c("X", "Y", "Z")))
#' sl <- streamline(points = pts)
#' b <- bundle(streamlines = list(sl))
#' b@n_streamlines  # 1
#' b@bundle_attributes  # NULL (no bundle-level attributes)
#'
#' # bundle_data is stored
#' b2 <- bundle(
#'   streamlines = list(sl),
#'   bundle_data = list(subject = "sub-01")
#' )
#' b2@bundle_data$subject  # "sub-01"
#'
#' # format(), print(), length() and indexing methods
#' format(b2)
#' print(b2)
#' length(b2)   # 1
#' b2[[1]]      # first streamline
#'
#' # subsetting preserves bundle_data
#' pts2 <- matrix(runif(15), ncol = 3, dimnames = list(NULL, c("X", "Y", "Z")))
#' sl2 <- streamline(points = pts2)
#' b3 <- bundle(streamlines = list(sl, sl2), bundle_data = list(subject = "sub-01"))
#' b3[1]@n_streamlines  # 1, bundle_data preserved
bundle <- S7::new_class(
  name = "bundle",
  package = "fiber",
  properties = list(
    streamlines = S7::class_list,
    bundle_data = S7::class_list,
    n_streamlines = S7::new_property(
      class = S7::class_integer,
      getter = function(self) length(self@streamlines)
    ),
    bundle_attributes = S7::new_property(
      class = S7::class_character,
      getter = function(self) names(self@bundle_data)
    )
  ),
  validator = function(self) {
    bad <- !vapply(
      self@streamlines,
      function(x) S7::S7_inherits(x, streamline),
      logical(1L)
    )
    if (any(bad)) {
      return(
        "All elements of @streamlines must be <fiber::streamline> objects."
      )
    }
    NULL
  }
)

# ---- predicate --------------------------------------------------------------

#' Test whether an object is a bundle
#'
#' @param x An object.
#' @returns `TRUE` if `x` is of class [bundle], otherwise `FALSE`.
#' @export
#' @examples
#' pts <- matrix(runif(15), ncol = 3, dimnames = list(NULL, c("X", "Y", "Z")))
#' sl <- streamline(points = pts)
#' b <- bundle(streamlines = list(sl))
#' is_bundle(b)   # TRUE
#' is_bundle(sl)  # FALSE
is_bundle <- function(x) S7::S7_inherits(x, bundle)

# ---- format / print ---------------------------------------------------------

S7::method(format, bundle) <- function(x, ...) {
  n <- x@n_streamlines
  if (n == 0L) {
    return("<bundle [0 streamlines]>")
  }
  npts <- vapply(x@streamlines, n_points, integer(1L))
  pd_keys <- unique(unlist(lapply(x@streamlines, function(s) {
    names(s@point_data)
  })))
  sld_keys <- unique(unlist(lapply(x@streamlines, function(s) {
    names(s@streamline_data)
  })))
  bd_keys <- names(x@bundle_data)
  pd_str <- if (length(pd_keys) > 0L) {
    paste0(" | point: ", paste(pd_keys, collapse = ", "))
  } else {
    ""
  }
  sld_str <- if (length(sld_keys) > 0L) {
    paste0(" | streamline: ", paste(sld_keys, collapse = ", "))
  } else {
    ""
  }
  bd_str <- if (length(bd_keys) > 0L) {
    paste0(" | bundle: ", paste(bd_keys, collapse = ", "))
  } else {
    ""
  }
  paste0(
    "<bundle [",
    n,
    " streamlines | ",
    min(npts),
    "\u2013",
    max(npts),
    " pts/streamline]",
    pd_str,
    sld_str,
    bd_str,
    ">"
  )
}

S7::method(print, bundle) <- function(x, ...) {
  cat(format(x, ...), "\n")
  invisible(x)
}

# ---- length / indexing for bundle -------------------------------------------

S7::method(length, bundle) <- function(x) x@n_streamlines

S7::method(`[[`, bundle) <- function(x, i, ...) x@streamlines[[i]]

S7::method(`[`, bundle) <- function(x, i, j, ..., drop = TRUE) {
  bundle(streamlines = x@streamlines[i], bundle_data = x@bundle_data)
}

# ---- S7 class: bundle_set ---------------------------------------------------

#' Bundle set S7 class
#'
#' @description
#' A `bundle_set` is a **named collection of [bundle] objects**, designed for
#' multi-subject or multi-session studies where each element represents one
#' subject's (or session's) tractogram. It stores two compartments:
#'
#' - `@bundles` — a *named* list of [bundle] objects. Names typically encode
#'   subject or session identifiers (e.g. `"sub-01"`, `"sub-02"`).
#' - `@set_data` — a named list of set-level metadata (arbitrary R objects,
#'   e.g. study name, atlas used, acquisition protocol).
#'
#' @param bundles A named list of [bundle] objects.
#' @param set_data A named list of set-level metadata.
#' @prop n_bundles An integer scalar giving the number of bundles in the set
#'   (read-only).
#' @prop bundle_names A character vector of the names of the bundles
#'   (read-only).
#' @prop set_attributes A character vector of the names of the set-level
#'   attributes (read-only).
#'
#' @returns A `bundle_set` S7 object.
#' @section Methods for standard generics:
#' The following methods are defined for `bundle_set` objects:
#' - `format(x, ...)`: Returns a compact character string.
#' - `print(x, ...)`: Prints the formatted string to the console and
#'   invisibly returns `x`.
#' - `length(x)`: Returns the number of bundles.
#' - `x[[i]]`: Extracts the `i`-th (or named) [bundle] from the set.
#' - `x[i]`: Returns a new [bundle_set] containing only the selected bundles,
#'   preserving `@set_data`.
#' - `names(x)`: Returns the names of the bundles.
#' @export
#' @examples
#' pts <- matrix(runif(15), ncol = 3, dimnames = list(NULL, c("X", "Y", "Z")))
#' b1 <- bundle(streamlines = list(streamline(points = pts)),
#'              bundle_data = list(subject = "sub-01"))
#' b2 <- bundle(streamlines = list(streamline(points = pts)),
#'              bundle_data = list(subject = "sub-02"))
#' bs <- bundle_set(bundles = list("sub-01" = b1, "sub-02" = b2))
#' bs@n_bundles      # 2
#' bs@bundle_names   # c("sub-01", "sub-02")
#' bs[["sub-01"]]    # first bundle
bundle_set <- S7::new_class(
  name = "bundle_set",
  package = "fiber",
  properties = list(
    bundles = S7::class_list,
    set_data = S7::class_list,
    n_bundles = S7::new_property(
      class = S7::class_integer,
      getter = function(self) length(self@bundles)
    ),
    bundle_names = S7::new_property(
      class = S7::class_character,
      getter = function(self) names(self@bundles) %||% character(0L)
    ),
    set_attributes = S7::new_property(
      class = S7::class_character,
      getter = function(self) names(self@set_data)
    )
  ),
  validator = function(self) {
    bad <- !vapply(
      self@bundles,
      function(x) S7::S7_inherits(x, bundle),
      logical(1L)
    )
    if (any(bad)) {
      return("All elements of @bundles must be <fiber::bundle> objects.")
    }
    if (length(self@bundles) > 0L && is.null(names(self@bundles))) {
      return("@bundles must be a named list.")
    }
    NULL
  }
)

# ---- predicate --------------------------------------------------------------

#' Test whether an object is a bundle_set
#'
#' @param x An object.
#' @returns `TRUE` if `x` is of class [bundle_set], otherwise `FALSE`.
#' @export
#' @examples
#' pts <- matrix(runif(15), ncol = 3, dimnames = list(NULL, c("X", "Y", "Z")))
#' b <- bundle(streamlines = list(streamline(points = pts)))
#' bs <- bundle_set(bundles = list("sub-01" = b))
#' is_bundle_set(bs)  # TRUE
#' is_bundle_set(b)   # FALSE
is_bundle_set <- function(x) S7::S7_inherits(x, bundle_set)

# ---- format / print ---------------------------------------------------------

S7::method(format, bundle_set) <- function(x, ...) {
  n <- x@n_bundles
  if (n == 0L) {
    return("<bundle_set [0 bundles]>")
  }
  nms <- x@bundle_names
  nm_preview <- if (n <= 3L) {
    paste(nms, collapse = ", ")
  } else {
    paste0(paste(nms[1:3], collapse = ", "), ", \u2026 (", n - 3L, " more)")
  }
  sd_keys <- names(x@set_data)
  sd_str <- if (length(sd_keys) > 0L) {
    paste0(" | set: ", paste(sd_keys, collapse = ", "))
  } else {
    ""
  }
  n_sls <- vapply(x@bundles, function(b) b@n_streamlines, integer(1L))
  paste0(
    "<bundle_set [",
    n,
    " bundles | ",
    min(n_sls),
    "\u2013",
    max(n_sls),
    " streamlines/bundle]: ",
    nm_preview,
    sd_str,
    ">"
  )
}

S7::method(print, bundle_set) <- function(x, ...) {
  cat(format(x, ...), "\n")
  invisible(x)
}

# ---- length / names / indexing ----------------------------------------------

S7::method(length, bundle_set) <- function(x) x@n_bundles

S7::method(names, bundle_set) <- function(x) x@bundle_names

S7::method(`[[`, bundle_set) <- function(x, i, ...) x@bundles[[i]]

S7::method(`[`, bundle_set) <- function(x, i, j, ..., drop = TRUE) {
  bundle_set(bundles = x@bundles[i], set_data = x@set_data)
}

library(fiber)

# ---- helpers -----------------------------------------------------------------

pts <- matrix(
  c(1, 2, 3, 4, 5, 6, 7, 8, 9),
  ncol = 3,
  dimnames = list(NULL, c("X", "Y", "Z"))
)
sl <- streamline(pts)
b1 <- bundle(list(sl), bundle_data = list(subject = "sub-01"))
b2 <- bundle(list(sl, sl), bundle_data = list(subject = "sub-02"))

# ---- bundle_set constructor -------------------------------------------------

bs <- bundle_set(bundles = list("sub-01" = b1, "sub-02" = b2))
expect_true(is_bundle_set(bs))
expect_equal(bs@n_bundles, 2L)
expect_equal(bs@bundle_names, c("sub-01", "sub-02"))

# empty bundle_set is allowed
bs0 <- bundle_set(bundles = list())
expect_true(is_bundle_set(bs0))
expect_equal(bs0@n_bundles, 0L)

# unnamed non-empty list -> error
expect_error(bundle_set(bundles = list(b1)))

# non-bundle element -> error
expect_error(bundle_set(bundles = list("x" = sl)))

# set_data is stored
bs_meta <- bundle_set(
  bundles = list("sub-01" = b1),
  set_data = list(study = "phantom")
)
expect_equal(bs_meta@set_data[["study"]], "phantom")
expect_equal(bs_meta@set_attributes, "study")

# ---- is_bundle_set -----------------------------------------------------------

expect_true(is_bundle_set(bs))
expect_false(is_bundle_set(b1))
expect_false(is_bundle_set(sl))

# ---- format.bundle_set -------------------------------------------------------

f0 <- format(bs0)
expect_true(grepl("0 bundles", f0))

f2 <- format(bs)
expect_true(grepl("2 bundles", f2))
expect_true(grepl("sub-01", f2))

# set_data keys appear in format
f_meta <- format(bs_meta)
expect_true(grepl("set: study", f_meta))

# ---- print.bundle_set --------------------------------------------------------

expect_stdout(print(bs), "bundles")

# ---- length / names ----------------------------------------------------------

expect_equal(length(bs), 2L)
expect_equal(names(bs), c("sub-01", "sub-02"))

# ---- indexing ----------------------------------------------------------------

# [[ by name
expect_true(is_bundle(bs[["sub-01"]]))

# [[ by position
expect_true(is_bundle(bs[[2L]]))

# [ returns a bundle_set preserving set_data
bs_sub <- bs_meta["sub-01"]
expect_true(is_bundle_set(bs_sub))
expect_equal(bs_sub@set_data[["study"]], "phantom")
expect_equal(bs_sub@n_bundles, 1L)

# ---- as_bundle_set -----------------------------------------------------------

# identity
expect_true(is_bundle_set(as_bundle_set(bs)))

# wrap a bundle
bs_from_b <- as_bundle_set(b1, name = "sub-01")
expect_true(is_bundle_set(bs_from_b))
expect_equal(bs_from_b@bundle_names, "sub-01")
expect_equal(bs_from_b@n_bundles, 1L)

# unknown class -> error
expect_error(as_bundle_set("not a bundle"))

# ---- bind_bundle_sets --------------------------------------------------------

bs1 <- bundle_set(list("sub-01" = b1))
bs2 <- bundle_set(list("sub-02" = b2))

# two bundle_sets
bs_all <- bind_bundle_sets(bs1, bs2)
expect_equal(bs_all@n_bundles, 2L)
expect_equal(sort(bs_all@bundle_names), c("sub-01", "sub-02"))

# named bare bundles
bs_named <- bind_bundle_sets("sub-01" = b1, "sub-02" = b2)
expect_equal(bs_named@n_bundles, 2L)

# mix of bundle_set and named bundle
bs_mix <- bind_bundle_sets(bs1, "sub-02" = b2)
expect_equal(bs_mix@n_bundles, 2L)

# set_data inherited from first bundle_set
bs1_meta <- bundle_set(list("sub-01" = b1), set_data = list(study = "test"))
bs_sd <- bind_bundle_sets(bs1_meta, bs2)
expect_equal(bs_sd@set_data[["study"]], "test")

# set_data override
bs_sd2 <- bind_bundle_sets(bs1_meta, bs2, set_data = list(study = "override"))
expect_equal(bs_sd2@set_data[["study"]], "override")

# unnamed bundle -> error
expect_error(bind_bundle_sets(b1))

# duplicate names -> error
expect_error(bind_bundle_sets(bs1, bs1))

# no args -> error
expect_error(bind_bundle_sets())

# wrong type -> error
expect_error(bind_bundle_sets("x" = "not_a_bundle"))

# Generated by cpp11: do not edit by hand

hausdorff_distance_cpp <- function(mat_a, mat_b) {
  .Call(`_fiber_hausdorff_distance_cpp`, mat_a, mat_b)
}

pairwise_hausdorff_cpp <- function(mats) {
  .Call(`_fiber_pairwise_hausdorff_cpp`, mats)
}

#' @include streamline.R bundle.R
NULL

# ---- as_streamline ----------------------------------------------------------

#' Coerce an object to a streamline
#'
#' `as_streamline()` converts a supported object into a [streamline].
#'
#' Currently supported input classes:
#' - [streamline]: returned unchanged.
#' - `dwiFiber` (from \pkg{dti}): the object must contain **exactly one**
#'   fiber. For multi-fiber objects use [as_bundle()] instead.
#'
#' @param x An object to coerce.
#' @param ... Additional arguments (currently unused).
#' @returns A [streamline] object.
#' @seealso [as_bundle()], [as_dwifiber()]
#' @export
#' @examples
#' pts <- matrix(runif(15), ncol = 3, dimnames = list(NULL, c("X", "Y", "Z")))
#' sl <- streamline(points = pts)
#' identical(as_streamline(sl), sl)  # TRUE — identity coercion
as_streamline <- S7::new_generic("as_streamline", "x")

# --- method: streamline (identity) ------------------------------------------

#' [as_streamline()] method for `streamline` objects
#'
#' @param x A [streamline] object.
#' @param ... Additional arguments (currently unused).
#' @returns `x` unchanged.
#' @seealso [as_streamline()]
#' @name as_streamline-fiber-streamline-method
#' @aliases as_streamline,fiber::streamline-method
#' @usage NULL
S7::method(as_streamline, streamline) <- function(x, ...) x

# --- method: bundle ----------------------------------------------------------

#' [as_streamline()] method for `bundle` objects
#'
#' @param x A [bundle] object containing exactly one streamline.
#' @param ... Additional arguments (currently unused).
#' @returns The sole [streamline] inside `x`.
#' @seealso [as_streamline()]
#' @name as_streamline-fiber-bundle-method
#' @aliases as_streamline,fiber::bundle-method
#' @usage NULL
S7::method(as_streamline, bundle) <- function(x, ...) {
  if (x@n_streamlines != 1L) {
    cli::cli_abort(c(
      "Cannot coerce a {.cls bundle} with {x@n_streamlines} streamlines \\
       to a single {.cls streamline}.",
      "i" = "Use {.code x[[i]]} to extract a specific streamline."
    ))
  }
  x[[1L]]
}

# --- method: class_any (handles dwiFiber + unknown classes) -----------------

S7::method(as_streamline, S7::class_any) <- function(x, ...) {
  if (inherits(x, "dwiFiber")) {
    b <- .dwifiber_to_bundle(x)
    if (b@n_streamlines != 1L) {
      cli::cli_abort(c(
        "Cannot coerce a {.cls dwiFiber} with {b@n_streamlines} fibers \\
         to a single {.cls streamline}.",
        "i" = "Use {.fn as_bundle} instead."
      ))
    }
    return(b[[1L]])
  }
  cli::cli_abort(
    "Don't know how to coerce an object of class {.cls {class(x)[1L]}} \\
     to a {.cls streamline}."
  )
}

# ---- as_bundle --------------------------------------------------------------

#' Coerce an object to a bundle
#'
#' `as_bundle()` converts a supported object into a [bundle].
#'
#' Currently supported input classes:
#' - [streamline]: wrapped in a single-element [bundle] (lossless).
#' - [bundle]: returned unchanged.
#' - `dwiFiber` (from \pkg{dti}): each fiber becomes a [streamline]. The
#'   per-point direction vectors (columns 4–6 of `@fibers`) are stored as
#'   `@point_data$direction_x`, `@point_data$direction_y`, and
#'   `@point_data$direction_z`. Tracking metadata (`method`, `minfa`,
#'   `maxangle`) are stored in `@bundle_data`.
#'
#' @param x An object to coerce.
#' @param ... Additional arguments (currently unused).
#' @returns A [bundle] object.
#' @seealso [as_streamline()], [as_dwifiber()]
#' @export
#' @examples
#' pts <- matrix(runif(15), ncol = 3, dimnames = list(NULL, c("X", "Y", "Z")))
#' sl <- streamline(points = pts)
#' b  <- as_bundle(sl)
#' b@n_streamlines  # 1
as_bundle <- S7::new_generic("as_bundle", "x")

# --- method: streamline (wrap) -----------------------------------------------

#' [as_bundle()] method for `streamline` objects
#'
#' @param x A [streamline] object.
#' @param ... Additional arguments (currently unused).
#' @returns A [bundle] containing `x` as its sole streamline.
#' @seealso [as_bundle()]
#' @name as_bundle-fiber-streamline-method
#' @aliases as_bundle,fiber::streamline-method
#' @usage NULL
S7::method(as_bundle, streamline) <- function(x, ...) {
  bundle(streamlines = list(x))
}

# --- method: bundle (identity) -----------------------------------------------

#' [as_bundle()] method for `bundle` objects
#'
#' @param x A [bundle] object.
#' @param ... Additional arguments (currently unused).
#' @returns `x` unchanged.
#' @seealso [as_bundle()]
#' @name as_bundle-fiber-bundle-method
#' @aliases as_bundle,fiber::bundle-method
#' @usage NULL
S7::method(as_bundle, bundle) <- function(x, ...) x

# --- method: class_any (handles dwiFiber + unknown classes) ------------------

S7::method(as_bundle, S7::class_any) <- function(x, ...) {
  if (inherits(x, "dwiFiber")) {
    return(.dwifiber_to_bundle(x))
  }
  cli::cli_abort(
    "Don't know how to coerce an object of class {.cls {class(x)[1L]}} \\
     to a {.cls bundle}."
  )
}

# ---- as_dwifiber ------------------------------------------------------------

#' Coerce a streamline or bundle to a `dwiFiber` object
#'
#' `as_dwifiber()` converts a [streamline] or [bundle] to the S4 class
#' `dwiFiber` from the \pkg{dti} package.
#'
#' Per-point direction vectors are taken from `@point_data$direction_x`,
#' `@point_data$direction_y`, and `@point_data$direction_z` when
#' present; otherwise they are estimated via finite differences of the
#' coordinates (forward difference at the first point, backward difference at
#' the last, central differences in between), then unit-normalised.
#'
#' Bundle-level metadata stored in `@bundle_data` under the keys `method`,
#' `minfa`, `maxangle`, `ddim`, `ddim0`, `voxelext`, `orientation`, `rotation`,
#' `level`, and `source` are transferred to the corresponding `dwiFiber` / `dwi`
#' slots when present. MRI-acquisition metadata that cannot be recovered from a
#' `fiber` object (gradient directions, b-values, etc.) are filled with neutral
#' placeholders.
#'
#' @param x A [streamline] or [bundle] object.
#' @param ... Additional arguments (currently unused).
#' @returns An S4 object of class `dwiFiber` (from \pkg{dti}).
#' @seealso [as_streamline()], [as_bundle()]
#' @export
#' @examples
#' if (requireNamespace("dti", quietly = TRUE)) {
#'   pts <- matrix(runif(15), ncol = 3, dimnames = list(NULL, c("X", "Y", "Z")))
#'   sl  <- streamline(points = pts)
#'   b   <- bundle(streamlines = list(sl))
#'   dfi <- as_dwifiber(b)
#'   class(dfi)  # "dwiFiber"
#' }
as_dwifiber <- S7::new_generic("as_dwifiber", "x")

# --- method: streamline ------------------------------------------------------

#' [as_dwifiber()] method for `streamline` objects
#'
#' @param x A [streamline] object.
#' @param ... Additional arguments (currently unused).
#' @returns An S4 `dwiFiber` object.
#' @seealso [as_dwifiber()]
#' @name as_dwifiber-fiber-streamline-method
#' @aliases as_dwifiber,fiber::streamline-method
#' @usage NULL
S7::method(as_dwifiber, streamline) <- function(x, ...) {
  as_dwifiber(bundle(streamlines = list(x)), ...)
}

# --- method: bundle ----------------------------------------------------------

#' [as_dwifiber()] method for `bundle` objects
#'
#' @param x A [bundle] object.
#' @param ... Additional arguments (currently unused).
#' @returns An S4 `dwiFiber` object.
#' @seealso [as_dwifiber()]
#' @name as_dwifiber-fiber-bundle-method
#' @aliases as_dwifiber,fiber::bundle-method
#' @usage NULL
S7::method(as_dwifiber, bundle) <- function(x, ...) {
  if (!requireNamespace("dti", quietly = TRUE)) {
    cli::cli_abort(c(
      "Package {.pkg dti} is required to create a {.cls dwiFiber} object.",
      "i" = "Install it with {.code install.packages(\"dti\")}."
    ))
  }
  .bundle_to_dwifiber(x)
}

# ---- internal helpers -------------------------------------------------------

# Convert a dwiFiber S4 object to a fiber::bundle.
.dwifiber_to_bundle <- function(x) {
  fibers_mat <- x@fibers
  startind <- x@startind
  n_fibers <- length(startind)
  n_rows <- nrow(fibers_mat)
  endind <- c(startind[-1L] - 1L, n_rows)

  streamlines <- vector("list", n_fibers)
  for (i in seq_len(n_fibers)) {
    rows <- seq.int(startind[i], endind[i])
    pts <- fibers_mat[rows, 1:3, drop = FALSE]
    colnames(pts) <- c("X", "Y", "Z")
    dirs <- fibers_mat[rows, 4:6, drop = FALSE]
    colnames(dirs) <- NULL
    streamlines[[i]] <- streamline(
      points = pts,
      point_data = list(
        direction_x = dirs[, 1],
        direction_y = dirs[, 2],
        direction_z = dirs[, 3]
      )
    )
  }

  bundle(
    streamlines = streamlines,
    bundle_data = list(
      method = x@method,
      minfa = x@minfa,
      maxangle = x@maxangle,
      ddim = x@ddim,
      ddim0 = x@ddim0,
      voxelext = x@voxelext,
      orientation = x@orientation,
      rotation = x@rotation,
      level = x@level,
      source = x@source
    )
  )
}

# Estimate unit-normalised tangent directions from a coordinate matrix using
# forward/central/backward finite differences.
.compute_directions <- function(pts) {
  n <- nrow(pts)
  dirs <- matrix(0, nrow = n, ncol = 3L)
  if (n == 1L) {
    return(dirs)
  }
  dirs[1L, ] <- pts[2L, ] - pts[1L, ] # forward
  dirs[n, ] <- pts[n, ] - pts[n - 1L, ] # backward
  if (n > 2L) {
    dirs[2L:(n - 1L), ] <- pts[3L:n, ] - pts[1L:(n - 2L), ] # central
  }
  norms <- sqrt(rowSums(dirs^2L))
  norms[norms == 0] <- 1
  dirs / norms
}

# Convert a fiber::bundle to a dwiFiber S4 object.
.bundle_to_dwifiber <- function(x) {
  n_fibers <- x@n_streamlines
  fibers_list <- vector("list", n_fibers)
  startind <- integer(n_fibers)
  row_cursor <- 1L
  bd <- x@bundle_data

  for (i in seq_len(n_fibers)) {
    sl <- x@streamlines[[i]]
    pts <- sl@points[, c("X", "Y", "Z"), drop = FALSE]
    n_pts <- nrow(pts)
    dirs <- if (
      all(
        c("direction_x", "direction_y", "direction_z") %in% sl@point_attributes
      )
    ) {
      cbind(
        sl@point_data[["direction_x"]],
        sl@point_data[["direction_y"]],
        sl@point_data[["direction_z"]]
      )
    } else {
      .compute_directions(pts)
    }
    fibers_list[[i]] <- cbind(pts, dirs)
    startind[i] <- row_cursor
    row_cursor <- row_cursor + n_pts
  }

  fibers_mat <- do.call(rbind, fibers_list)
  colnames(fibers_mat) <- c("x", "y", "z", "dx", "dy", "dz")

  methods::new(
    "dwiFiber",
    fibers = fibers_mat,
    startind = as.integer(startind),
    roimask = as.raw(0), # placeholder; dwiFiber requires a roimask slot but we have no equivalent data to fill it with
    method = as.character(bd[["method"]] %||% "unknown"),
    minfa = as.numeric(bd[["minfa"]] %||% 0),
    maxangle = as.numeric(bd[["maxangle"]] %||% 0),
    # dwi superclass slots — filled with neutral placeholders when absent
    call = as.list(call("as_dwifiber")),
    gradient = bd[["gradient"]] %||% matrix(0, nrow = 3L, ncol = 0L),
    bvalue = bd[["bvalue"]] %||% numeric(0L),
    btb = bd[["btb"]] %||% matrix(0, nrow = 6L, ncol = 0L),
    mask = bd[["mask"]] %||% array(FALSE, dim = c(0L, 0L, 0L)),
    ngrad = bd[["ngrad"]] %||% 0L,
    s0ind = bd[["s0ind"]] %||% integer(0L),
    replind = bd[["replind"]] %||% integer(0L),
    ddim = as.integer(bd[["ddim"]] %||% c(0L, 0L, 0L)),
    ddim0 = as.integer(bd[["ddim0"]] %||% c(0L, 0L, 0L)),
    xind = as.integer(bd[["xind"]] %||% integer(0L)),
    yind = as.integer(bd[["yind"]] %||% integer(0L)),
    zind = as.integer(bd[["zind"]] %||% integer(0L)),
    voxelext = as.numeric(bd[["voxelext"]] %||% c(1, 1, 1)),
    level = as.numeric(bd[["level"]] %||% 0),
    orientation = as.integer(bd[["orientation"]] %||% c(0L, 2L, 5L)),
    rotation = bd[["rotation"]] %||% diag(3),
    source = as.character(bd[["source"]] %||% "")
  )
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/streamline.R
\name{streamline}
\alias{streamline}
\title{Streamline S7 class}
\usage{
streamline(points = NULL, point_data = list(), streamline_data = list())
}
\arguments{
\item{points}{A numeric matrix with columns \code{"X"}, \code{"Y"}, and \code{"Z"}.}

\item{point_data}{A named list of per-point numeric vectors.}

\item{streamline_data}{A named list of per-streamline numeric scalars.}
}
\value{
A \code{streamline} S7 object.
}
\description{
A \code{streamline} represents a single fibre tract. It stores three data
compartments that mirror the conceptual levels found in tractography file
formats:
\itemize{
\item \verb{@points} — an \eqn{n \times 3} numeric matrix whose columns are named
\code{"X"}, \code{"Y"}, and \code{"Z"}, holding the ordered 3-D coordinates of the
\eqn{n} points along the tract.
\item \verb{@point_data} — a named list of numeric vectors, each of length \eqn{n},
holding per-point scalar attributes (e.g. fractional anisotropy sampled
at every point).
\item \verb{@streamline_data} — a named list of numeric scalars (length-1 vectors)
holding per-streamline attributes (e.g. a tract-level weight or mean FA).
}
}
\section{Methods for standard generics}{

The following methods are defined for \code{streamline} objects:
\itemize{
\item \code{format(x, ...)}: Returns a compact character string such as
\verb{<streamline [10 pts] | point: FA>}.
\item \code{print(x, ...)}: Prints the formatted string to the console and
invisibly returns \code{x}.
}
}

\examples{
# Create a streamline with 5 points and some attributes
sl <- streamline(
  points = matrix(
    c(0, 0, 0,
      1, 0, 0,
      1, 1, 0,
      1, 1, 1,
      0, 1, 1),
    ncol = 3,
    byrow = TRUE,
    dimnames = list(NULL, c("X", "Y", "Z"))
  ),
  point_data = list(FA = c(0.5, 0.6, 0.7, 0.8, 0.9)),
  streamline_data = list(mean_FA = 0.7)
)
sl@n_points  # 5
sl@point_attributes  # "FA"
sl@streamline_attributes  # "mean_FA"

# format() and print() methods
format(sl)  # "<streamline [5 pts] | point: FA | streamline: mean_FA>"
print(sl)
}
\section{Additional properties}{

\describe{
\item{\code{@n_points}}{An integer scalar giving the number of points in the streamline (read-only).}

\item{\code{@point_attributes}}{A character vector of the names of the per-point attributes (read-only).}

\item{\code{@streamline_attributes}}{A character vector of the names of the per-streamline attributes (read-only).}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coerce.R
\name{as_dwifiber}
\alias{as_dwifiber}
\title{Coerce a streamline or bundle to a \code{dwiFiber} object}
\usage{
as_dwifiber(x, ...)
}
\arguments{
\item{x}{A \link{streamline} or \link{bundle} object.}

\item{...}{Additional arguments (currently unused).}
}
\value{
An S4 object of class \code{dwiFiber} (from \pkg{dti}).
}
\description{
\code{as_dwifiber()} converts a \link{streamline} or \link{bundle} to the S4 class
\code{dwiFiber} from the \pkg{dti} package.
}
\details{
Per-point direction vectors are taken from \verb{@point_data$direction_x},
\verb{@point_data$direction_y}, and \verb{@point_data$direction_z} when
present; otherwise they are estimated via finite differences of the
coordinates (forward difference at the first point, backward difference at
the last, central differences in between), then unit-normalised.

Bundle-level metadata stored in \verb{@bundle_data} under the keys \code{method},
\code{minfa}, \code{maxangle}, \code{ddim}, \code{ddim0}, \code{voxelext}, \code{orientation}, \code{rotation},
\code{level}, and \code{source} are transferred to the corresponding \code{dwiFiber} / \code{dwi}
slots when present. MRI-acquisition metadata that cannot be recovered from a
\code{fiber} object (gradient directions, b-values, etc.) are filled with neutral
placeholders.
}
\examples{
if (requireNamespace("dti", quietly = TRUE)) {
  pts <- matrix(runif(15), ncol = 3, dimnames = list(NULL, c("X", "Y", "Z")))
  sl  <- streamline(points = pts)
  b   <- bundle(streamlines = list(sl))
  dfi <- as_dwifiber(b)
  class(dfi)  # "dwiFiber"
}
}
\seealso{
\code{\link[=as_streamline]{as_streamline()}}, \code{\link[=as_bundle]{as_bundle()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parametrize.R
\name{bind_bundles}
\alias{bind_bundles}
\title{Combine streamlines and/or bundles into a single bundle}
\usage{
bind_bundles(..., bundle_data = NULL)
}
\arguments{
\item{...}{One or more \link{streamline} or \link{bundle} objects.}

\item{bundle_data}{A named list of bundle-level metadata to attach to the
resulting \link{bundle}. Defaults to an empty list (or the \code{bundle_data} of
the first \link{bundle} input if one is present and \code{bundle_data} is not
supplied).}
}
\value{
A \link{bundle} containing all input streamlines.
}
\description{
Accepts any mix of \link{streamline} and \link{bundle} objects. All streamlines are
collected into a flat list and wrapped in a new \link{bundle}. \code{bundle_data}
from the first \link{bundle} argument (if any) is preserved; pass your own via
the \code{bundle_data} argument to override.
}
\examples{
pts <- matrix(runif(15), ncol = 3, dimnames = list(NULL, c("X", "Y", "Z")))
sl1 <- streamline(points = pts)
sl2 <- streamline(points = pts)
b1 <- bundle(streamlines = list(sl1))
b2 <- bundle(streamlines = list(sl2))

# combine two bundles
b_all <- bind_bundles(b1, b2)
b_all@n_streamlines  # 2

# mix a bundle and a loose streamline
b_mixed <- bind_bundles(b1, sl2)
b_mixed@n_streamlines  # 2
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bundle.R
\name{bundle_set}
\alias{bundle_set}
\title{Bundle set S7 class}
\usage{
bundle_set(bundles = list(), set_data = list())
}
\arguments{
\item{bundles}{A named list of \link{bundle} objects.}

\item{set_data}{A named list of set-level metadata.}
}
\value{
A \code{bundle_set} S7 object.
}
\description{
A \code{bundle_set} is a \strong{named collection of \link{bundle} objects}, designed for
multi-subject or multi-session studies where each element represents one
subject's (or session's) tractogram. It stores two compartments:
\itemize{
\item \verb{@bundles} — a \emph{named} list of \link{bundle} objects. Names typically encode
subject or session identifiers (e.g. \code{"sub-01"}, \code{"sub-02"}).
\item \verb{@set_data} — a named list of set-level metadata (arbitrary R objects,
e.g. study name, atlas used, acquisition protocol).
}
}
\section{Methods for standard generics}{

The following methods are defined for \code{bundle_set} objects:
\itemize{
\item \code{format(x, ...)}: Returns a compact character string.
\item \code{print(x, ...)}: Prints the formatted string to the console and
invisibly returns \code{x}.
\item \code{length(x)}: Returns the number of bundles.
\item \code{x[[i]]}: Extracts the \code{i}-th (or named) \link{bundle} from the set.
\item \code{x[i]}: Returns a new \link{bundle_set} containing only the selected bundles,
preserving \verb{@set_data}.
\item \code{names(x)}: Returns the names of the bundles.
}
}

\examples{
pts <- matrix(runif(15), ncol = 3, dimnames = list(NULL, c("X", "Y", "Z")))
b1 <- bundle(streamlines = list(streamline(points = pts)),
             bundle_data = list(subject = "sub-01"))
b2 <- bundle(streamlines = list(streamline(points = pts)),
             bundle_data = list(subject = "sub-02"))
bs <- bundle_set(bundles = list("sub-01" = b1, "sub-02" = b2))
bs@n_bundles      # 2
bs@bundle_names   # c("sub-01", "sub-02")
bs[["sub-01"]]    # first bundle
}
\section{Additional properties}{

\describe{
\item{\code{@n_bundles}}{An integer scalar giving the number of bundles in the set
(read-only).}

\item{\code{@bundle_names}}{A character vector of the names of the bundles
(read-only).}

\item{\code{@set_attributes}}{A character vector of the names of the set-level
attributes (read-only).}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shape.R
\name{add_shape_descriptors}
\alias{add_shape_descriptors}
\title{Adds shape descriptors to a streamline or bundle}
\usage{
add_shape_descriptors(
  x,
  descriptors = c("euclidean_length", "curvilinear_length", "sinuosity", "curvature",
    "torsion")
)
}
\arguments{
\item{x}{A \link{streamline} or \link{bundle} object.}

\item{descriptors}{A character vector of shape descriptors to add. Defaults
to all available descriptors: \code{c("euclidean_length", "curvilinear_length", "sinuosity", "curvature", "torsion")}.}
}
\value{
An object of the same class as \code{x} with the specified shape descriptors
added to the \verb{@streamline_data} or \verb{@point_data} slots of each streamline.
}
\description{
\code{add_shape_descriptors()} is an S7 generic that computes a number of shape
descriptors for each \link{streamline} object and stores them in the
\verb{@streamline_data} or \verb{@point_data} slots as appropriate, with methods
available for the following classes:

\itemize{
\item \code{\link[=add_shape_descriptors-fiber-bundle-method]{fiber::bundle}}
\item \code{\link[=add_shape_descriptors-fiber-streamline-method]{fiber::streamline}}
}

This function provides a convenient way to compute shape descriptors and
attach them to \link{streamline} or \link{bundle} objects. See the documentation for
each individual shape descriptor function (e.g. \code{get_euclidean_length()},
\code{get_curvilinear_length()}, \code{get_sinuosity()}, \code{get_curvature()},
\code{get_torsion()}) for more details on how each descriptor is computed.
}
\examples{
# add multiple shape descriptors to a single streamline
pts <- matrix(runif(30), ncol = 3)
colnames(pts) <- c("X", "Y", "Z")
sl <- streamline(points = pts)
sl <- add_shape_descriptors(
  sl,
  descriptors = c("euclidean_length", "curvilinear_length", "sinuosity")
)
# add multiple shape descriptors to a bundle
sl1 <- streamline(points = pts)
pts2 <- matrix(runif(60), ncol = 3)
colnames(pts2) <- c("X", "Y", "Z")
sl2 <- streamline(points = pts2)
b <- bundle(streamlines = list(sl1, sl2))
b <- add_shape_descriptors(
  b,
  descriptors = c("euclidean_length", "curvilinear_length", "sinuosity")
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/streamline.R
\name{is_streamline}
\alias{is_streamline}
\title{Test whether an object is a streamline}
\usage{
is_streamline(x)
}
\arguments{
\item{x}{An object.}
}
\value{
\code{TRUE} if \code{x} is of class \link{streamline}, otherwise \code{FALSE}.
}
\description{
Test whether an object is a streamline
}
\examples{
pts <- matrix(runif(15), ncol = 3, dimnames = list(NULL, c("X", "Y", "Z")))
sl <- streamline(points = pts)
is_streamline(sl)     # TRUE
is_streamline(42)     # FALSE
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shape.R
\name{compute_hausdorff_distance-fiber-streamline-method}
\alias{compute_hausdorff_distance-fiber-streamline-method}
\alias{compute_hausdorff_distance,fiber::streamline-method}
\title{\code{\link[=compute_hausdorff_distance]{compute_hausdorff_distance()}} method for two \code{streamline} objects}
\arguments{
\item{x}{A \link{streamline} object.}

\item{y}{A \link{streamline} object.}
}
\value{
A non-negative numeric scalar equal to
\eqn{\max(d_H(x \to y),\, d_H(y \to x))}, where
\eqn{d_H(A \to B) = \max_{a \in A} \min_{b \in B} \|a - b\|_2} is the
directed Hausdorff distance. The core computation is performed in C++
via \code{hausdorff_distance_cpp()}.
}
\description{
\code{\link[=compute_hausdorff_distance]{compute_hausdorff_distance()}} method for two \code{streamline} objects
}
\examples{
pts1 <- matrix(runif(30), ncol = 3)
colnames(pts1) <- c("X", "Y", "Z")
pts2 <- matrix(runif(30), ncol = 3)
colnames(pts2) <- c("X", "Y", "Z")
sl1 <- streamline(points = pts1)
sl2 <- streamline(points = pts2)
compute_hausdorff_distance(sl1, sl2)
}
\seealso{
\code{\link[=compute_hausdorff_distance]{compute_hausdorff_distance()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bundle.R
\name{is_bundle_set}
\alias{is_bundle_set}
\title{Test whether an object is a bundle_set}
\usage{
is_bundle_set(x)
}
\arguments{
\item{x}{An object.}
}
\value{
\code{TRUE} if \code{x} is of class \link{bundle_set}, otherwise \code{FALSE}.
}
\description{
Test whether an object is a bundle_set
}
\examples{
pts <- matrix(runif(15), ncol = 3, dimnames = list(NULL, c("X", "Y", "Z")))
b <- bundle(streamlines = list(streamline(points = pts)))
bs <- bundle_set(bundles = list("sub-01" = b))
is_bundle_set(bs)  # TRUE
is_bundle_set(b)   # FALSE
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coerce.R
\name{as_bundle-fiber-streamline-method}
\alias{as_bundle-fiber-streamline-method}
\alias{as_bundle,fiber::streamline-method}
\title{\code{\link[=as_bundle]{as_bundle()}} method for \code{streamline} objects}
\arguments{
\item{x}{A \link{streamline} object.}

\item{...}{Additional arguments (currently unused).}
}
\value{
A \link{bundle} containing \code{x} as its sole streamline.
}
\description{
\code{\link[=as_bundle]{as_bundle()}} method for \code{streamline} objects
}
\seealso{
\code{\link[=as_bundle]{as_bundle()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shape.R
\name{get_torsion}
\alias{get_torsion}
\title{Torsion of a streamline}
\usage{
get_torsion(x)
}
\arguments{
\item{x}{A \link{streamline} object.}
}
\value{
A numeric vector of length \code{x@n_points} giving the torsion
\eqn{\tau(s)} at each sampled point along the streamline. Positive values
indicate right-handed twisting; negative values indicate left-handed
twisting; zero indicates a planar curve at that location.
}
\description{
\code{get_torsion()} is function that computes the torsion of a
\link{streamline} object. The torsion \eqn{\tau(s)} at each point
along the arc-length abscissa is computed using cubic smoothing
splines (4 degrees of freedom per component).
}
\examples{
pts <- matrix(runif(30), ncol = 3)
colnames(pts) <- c("X", "Y", "Z")
sl <- streamline(points = pts)
get_torsion(sl)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/streamline.R
\name{n_points}
\alias{n_points}
\title{Number of points in a streamline}
\usage{
n_points(x)
}
\arguments{
\item{x}{A \link{streamline} object.}
}
\value{
An integer scalar giving the number of points.
}
\description{
Number of points in a streamline
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parametrize.R
\name{reparametrize-fiber-streamline-method}
\alias{reparametrize-fiber-streamline-method}
\alias{reparametrize,fiber::streamline-method}
\title{\code{\link[=reparametrize]{reparametrize()}} method for \code{streamline} objects}
\arguments{
\item{x}{A \link{streamline} object.}

\item{n_points}{Number of equally-spaced arc-length points to use.
\itemize{
\item For a single \link{streamline}, defaults to \code{nrow(x@points)}.
\item For a \link{bundle}, defaults to the rounded mean number of points across
all streamlines.
}

Pass \code{NULL} to use these defaults explicitly.}
}
\value{
A \link{streamline} reparametrized onto the new grid. The returned
object has the same class as the input but with \verb{@points} resampled to
exactly \code{n_points} rows and all \verb{@point_data} vectors resampled
correspondingly via linear interpolation.
}
\description{
Resamples a single \link{streamline} onto a uniform arc-length grid. See
\code{\link[=reparametrize]{reparametrize()}} for the full parameter documentation.
}
\examples{
pts <- matrix(runif(30), ncol = 3)
colnames(pts) <- c("X", "Y", "Z")
sl <- streamline(points = pts, point_data = list(FA = runif(10)))
sl_reparam <- reparametrize(sl, n_points = 20)
sl_reparam@n_points  # 20
}
\seealso{
\code{\link[=reparametrize]{reparametrize()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coerce.R
\name{as_streamline}
\alias{as_streamline}
\title{Coerce an object to a streamline}
\usage{
as_streamline(x, ...)
}
\arguments{
\item{x}{An object to coerce.}

\item{...}{Additional arguments (currently unused).}
}
\value{
A \link{streamline} object.
}
\description{
\code{as_streamline()} converts a supported object into a \link{streamline}.
}
\details{
Currently supported input classes:
\itemize{
\item \link{streamline}: returned unchanged.
\item \code{dwiFiber} (from \pkg{dti}): the object must contain \strong{exactly one}
fiber. For multi-fiber objects use \code{\link[=as_bundle]{as_bundle()}} instead.
}
}
\examples{
pts <- matrix(runif(15), ncol = 3, dimnames = list(NULL, c("X", "Y", "Z")))
sl <- streamline(points = pts)
identical(as_streamline(sl), sl)  # TRUE — identity coercion
}
\seealso{
\code{\link[=as_bundle]{as_bundle()}}, \code{\link[=as_dwifiber]{as_dwifiber()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shape.R
\name{compute_hausdorff_distance}
\alias{compute_hausdorff_distance}
\title{Computes the Hausdorff distance between streamlines}
\usage{
compute_hausdorff_distance(x, y = NULL)
}
\arguments{
\item{x}{A \link{streamline} or \link{bundle} object.}

\item{y}{A \link{streamline} or \link{bundle} object, or \code{NULL} (default). When \code{NULL}
and \code{x} is a \link{bundle}, the pairwise distance matrix within \code{x} is
returned.}
}
\value{
\itemize{
\item A non-negative numeric scalar when both \code{x} and \code{y} are \link{streamline}s.
\item A \code{\link[stats:dist]{dist}} object of size \eqn{n} when \code{x} is a \link{bundle}
and \code{y} is \code{NULL} or a \link{bundle} (use \code{\link[=as.matrix]{as.matrix()}} to expand to a full
\eqn{n \times n} matrix).
\item A numeric vector of length \eqn{n} when \code{x} is a \link{bundle} and \code{y} is
a \link{streamline}.
}
}
\description{
\code{compute_hausdorff_distance()} is an S7 generic that computes the symmetric
Hausdorff distance between \link{streamline} objects based on their 3-D
coordinate matrices, with methods available for the following classes:

\itemize{
\item \code{ANY}
\item \code{\link[=compute_hausdorff_distance-fiber-bundle-method]{fiber::bundle}}
\item \code{\link[=compute_hausdorff_distance-fiber-streamline-method]{fiber::streamline}}
}

The four dispatch cases are:
\itemize{
\item \strong{\code{streamline} + \code{streamline}}: returns a single numeric scalar — the
symmetric Hausdorff distance between the two streamlines.
\item \strong{\code{bundle} + missing}: returns a symmetric numeric distance matrix of
dimension \eqn{n \times n}, where \eqn{n} is the number of streamlines in
the bundle, giving all pairwise Hausdorff distances.
\item \strong{\code{bundle} + \code{streamline}}: returns a numeric vector of length \eqn{n}
giving the Hausdorff distance from \code{y} to each streamline in \code{x}.
\item \strong{\code{bundle} + \code{bundle}}: returns a symmetric numeric distance matrix of
dimension \eqn{(n_x + n_y) \times (n_x + n_y)}, treating the
concatenation of all streamlines from \code{x} and \code{y} as one collection.
}
}
\examples{
pts1 <- matrix(runif(30), ncol = 3)
colnames(pts1) <- c("X", "Y", "Z")
sl1 <- streamline(points = pts1)
pts2 <- matrix(runif(30), ncol = 3)
colnames(pts2) <- c("X", "Y", "Z")
sl2 <- streamline(points = pts2)

# streamline x streamline -> scalar
compute_hausdorff_distance(sl1, sl2)

# bundle x missing -> pairwise dist object
b <- bundle(streamlines = list(sl1, sl2))
compute_hausdorff_distance(b)
as.matrix(compute_hausdorff_distance(b))

# bundle x streamline -> vector
compute_hausdorff_distance(b, sl1)

# bundle x bundle -> combined pairwise matrix
b2 <- bundle(streamlines = list(sl2))
compute_hausdorff_distance(b, b2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parametrize.R
\name{reparametrize-fiber-bundle-method}
\alias{reparametrize-fiber-bundle-method}
\alias{reparametrize,fiber::bundle-method}
\title{\code{\link[=reparametrize]{reparametrize()}} method for \code{bundle} objects}
\arguments{
\item{x}{A \link{bundle} object.}

\item{n_points}{Number of equally-spaced arc-length points to use.
\itemize{
\item For a single \link{streamline}, defaults to \code{nrow(x@points)}.
\item For a \link{bundle}, defaults to the rounded mean number of points across
all streamlines.
}

Pass \code{NULL} to use these defaults explicitly.}
}
\value{
A \link{bundle} reparametrized onto the new grid. Every streamline in the
returned bundle has exactly \code{n_points} rows in \verb{@points} (defaulting to
the rounded mean number of points across all streamlines when \code{n_points}
is \code{NULL}).
}
\description{
Resamples every \link{streamline} inside a \link{bundle} onto a common uniform
arc-length grid. See \code{\link[=reparametrize]{reparametrize()}} for the full parameter documentation.
}
\examples{
pts1 <- matrix(runif(30), ncol = 3)
colnames(pts1) <- c("X", "Y", "Z")
pts2 <- matrix(runif(60), ncol = 3)
colnames(pts2) <- c("X", "Y", "Z")
b <- bundle(streamlines = list(streamline(points = pts1),
                                streamline(points = pts2)))
b_reparam <- reparametrize(b, n_points = 15)
b_reparam[[1]]@n_points  # 15
b_reparam[[2]]@n_points  # 15
}
\seealso{
\code{\link[=reparametrize]{reparametrize()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shape.R
\name{compute_hausdorff_distance-fiber-bundle-method}
\alias{compute_hausdorff_distance-fiber-bundle-method}
\alias{compute_hausdorff_distance,fiber::bundle-method}
\title{\code{\link[=compute_hausdorff_distance]{compute_hausdorff_distance()}} method for \code{bundle} objects}
\arguments{
\item{x}{A \link{bundle} object.}

\item{y}{\code{NULL}, a \link{streamline}, or a \link{bundle}.}
}
\value{
\itemize{
\item A \code{\link[stats:dist]{dist}} object of size \code{x@n_streamlines} when \code{y} is
\code{NULL} or a \link{bundle}. The lower triangle stores all pairwise symmetric
Hausdorff distances (computed in C++). Use \code{\link[=as.matrix]{as.matrix()}} to obtain the
full \eqn{n \times n} matrix.
\item A numeric vector of length \code{x@n_streamlines} when \code{y} is a \link{streamline}.
}
}
\description{
Dispatches to one of three behaviours depending on \code{y}:
}
\details{
\itemize{
\item \code{y = NULL}: pairwise distances within \code{x} as a \code{\link[stats:dist]{dist}}
object of size \eqn{n}, computed in C++ via a single linear loop.
\item \code{y} is a \link{streamline}: numeric vector of distances from \code{y} to each
streamline in \code{x}.
\item \code{y} is a \link{bundle}: \code{\link[stats:dist]{dist}} object for the concatenation
of all streamlines from \code{x} and \code{y}.
}
}
\examples{
pts1 <- matrix(runif(30), ncol = 3)
colnames(pts1) <- c("X", "Y", "Z")
pts2 <- matrix(runif(30), ncol = 3)
colnames(pts2) <- c("X", "Y", "Z")
sl1 <- streamline(points = pts1)
sl2 <- streamline(points = pts2)
b <- bundle(streamlines = list(sl1, sl2))

# pairwise dist object (size 2)
compute_hausdorff_distance(b)
as.matrix(compute_hausdorff_distance(b))

# distances from sl1 to each streamline in b
compute_hausdorff_distance(b, sl1)
}
\seealso{
\code{\link[=compute_hausdorff_distance]{compute_hausdorff_distance()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bundle.R
\name{bundle}
\alias{bundle}
\title{Bundle S7 class}
\usage{
bundle(streamlines = list(), bundle_data = list())
}
\arguments{
\item{streamlines}{A list of \link{streamline} objects.}

\item{bundle_data}{A named list of bundle-level metadata.}
}
\value{
A \code{bundle} S7 object.
}
\description{
A \code{bundle} is an ordered collection of \link{streamline} objects representing a
tractogram or white-matter bundle. It stores two compartments:
\itemize{
\item \verb{@streamlines} — a list of \link{streamline} objects.
\item \verb{@bundle_data} — a named list of bundle-level metadata (arbitrary R
objects, e.g. the affine transform used during tracking).
}
}
\section{Methods for standard generics}{

The following methods are defined for \code{bundle} objects:
\itemize{
\item \code{format(x, ...)}: Returns a compact character string such as
\verb{<bundle [2 streamlines | 10–20 pts/streamline]>}.
\item \code{print(x, ...)}: Prints the formatted string to the console and
invisibly returns \code{x}.
\item \code{length(x)}: Returns the number of streamlines (equivalent to
\code{x@n_streamlines}).
\item \code{x[[i]]}: Extracts the \code{i}-th \link{streamline} from the bundle.
\item \code{x[i]}: Returns a new \link{bundle} containing only the selected
streamlines, preserving \verb{@bundle_data}.
}
}

\examples{
pts <- matrix(runif(15), ncol = 3, dimnames = list(NULL, c("X", "Y", "Z")))
sl <- streamline(points = pts)
b <- bundle(streamlines = list(sl))
b@n_streamlines  # 1
b@bundle_attributes  # NULL (no bundle-level attributes)

# bundle_data is stored
b2 <- bundle(
  streamlines = list(sl),
  bundle_data = list(subject = "sub-01")
)
b2@bundle_data$subject  # "sub-01"

# format(), print(), length() and indexing methods
format(b2)
print(b2)
length(b2)   # 1
b2[[1]]      # first streamline

# subsetting preserves bundle_data
pts2 <- matrix(runif(15), ncol = 3, dimnames = list(NULL, c("X", "Y", "Z")))
sl2 <- streamline(points = pts2)
b3 <- bundle(streamlines = list(sl, sl2), bundle_data = list(subject = "sub-01"))
b3[1]@n_streamlines  # 1, bundle_data preserved
}
\section{Additional properties}{

\describe{
\item{\code{@n_streamlines}}{An integer scalar giving the number of streamlines in the bundle (read-only).}

\item{\code{@bundle_attributes}}{A character vector of the names of the bundle-level attributes (read-only).}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fiber-package.R
\docType{package}
\name{fiber-package}
\alias{fiber}
\alias{fiber-package}
\title{fiber: S7 Data Structures for Diffusion MRI Tractography}
\description{
Provides three S7 classes — streamline, bundle, and bundle_set — for representing diffusion MRI tractography data in R, together with a concise set of methods for computing shape descriptors (arc-length, curvature, torsion, sinuosity), the Hausdorff distance between streamlines, arc-length reparametrization of streamlines and bundles onto uniform grids, combination of streamlines or bundles into a single bundle, combination of bundles from multiple subjects or sessions into a bundle_set, and coercion to and from the dwiFiber S4 class of the 'dti' package. See Dell'Acqua, F., Descoteaux, M. and Leemans, A. (2024) "Handbook of Diffusion MR Tractography" \doi{10.1016/C2018-0-02520-7} for more about the mathematical and computational underpinnings of diffusion MRI tractography.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/tractoverse/fiber}
  \item \url{https://tractoverse.github.io/fiber/}
  \item Report bugs at \url{https://github.com/tractoverse/fiber/issues}
}

}
\author{
\strong{Maintainer}: Aymeric Stamm \email{aymeric.stamm@cnrs.fr} (\href{https://orcid.org/0000-0002-8725-3654}{ORCID})

Authors:
\itemize{
  \item Aymeric Stamm \email{aymeric.stamm@cnrs.fr} (\href{https://orcid.org/0000-0002-8725-3654}{ORCID})
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coerce.R
\name{as_bundle}
\alias{as_bundle}
\title{Coerce an object to a bundle}
\usage{
as_bundle(x, ...)
}
\arguments{
\item{x}{An object to coerce.}

\item{...}{Additional arguments (currently unused).}
}
\value{
A \link{bundle} object.
}
\description{
\code{as_bundle()} converts a supported object into a \link{bundle}.
}
\details{
Currently supported input classes:
\itemize{
\item \link{streamline}: wrapped in a single-element \link{bundle} (lossless).
\item \link{bundle}: returned unchanged.
\item \code{dwiFiber} (from \pkg{dti}): each fiber becomes a \link{streamline}. The
per-point direction vectors (columns 4–6 of \verb{@fibers}) are stored as
\verb{@point_data$direction_x}, \verb{@point_data$direction_y}, and
\verb{@point_data$direction_z}. Tracking metadata (\code{method}, \code{minfa},
\code{maxangle}) are stored in \verb{@bundle_data}.
}
}
\examples{
pts <- matrix(runif(15), ncol = 3, dimnames = list(NULL, c("X", "Y", "Z")))
sl <- streamline(points = pts)
b  <- as_bundle(sl)
b@n_streamlines  # 1
}
\seealso{
\code{\link[=as_streamline]{as_streamline()}}, \code{\link[=as_dwifiber]{as_dwifiber()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parametrize.R
\name{bind_bundle_sets}
\alias{bind_bundle_sets}
\title{Combine bundles and/or bundle_sets into a single bundle_set}
\usage{
bind_bundle_sets(..., set_data = NULL)
}
\arguments{
\item{...}{Named \link{bundle} objects or \link{bundle_set} objects to combine. Each
bare \link{bundle} argument must be \strong{named} so that its label in the
resulting set is unambiguous.}

\item{set_data}{A named list of set-level metadata to attach to the
resulting \link{bundle_set}. Defaults to the \code{set_data} of the first
\link{bundle_set} input (if present) or an empty list.}
}
\value{
A \link{bundle_set} containing all input bundles.
}
\description{
Accepts any mix of named \link{bundle} objects (passed as \code{name = bundle}) or
\link{bundle_set} objects. All bundles are collected into a flat named list and
wrapped in a new \link{bundle_set}.
}
\examples{
pts <- matrix(runif(15), ncol = 3, dimnames = list(NULL, c("X", "Y", "Z")))
b1 <- bundle(streamlines = list(streamline(points = pts)))
b2 <- bundle(streamlines = list(streamline(points = pts)))

# two named bare bundles
bs <- bind_bundle_sets("sub-01" = b1, "sub-02" = b2)
bs@n_bundles   # 2
bs@bundle_names  # c("sub-01", "sub-02")

# combine two bundle_sets
bs1 <- bundle_set(list("sub-01" = b1))
bs2 <- bundle_set(list("sub-02" = b2))
bs_all <- bind_bundle_sets(bs1, bs2)
bs_all@n_bundles  # 2
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parametrize.R
\name{as_bundle_set}
\alias{as_bundle_set}
\title{Coerce an object to a bundle_set}
\usage{
as_bundle_set(x, ...)
}
\arguments{
\item{x}{An object to coerce.}

\item{...}{Additional arguments passed to methods (e.g. \code{name} for bundles).}
}
\value{
A \link{bundle_set} object.
}
\description{
\code{as_bundle_set()} converts a supported object into a \link{bundle_set}.
}
\details{
Currently supported input classes:
\itemize{
\item \link{bundle_set}: returned unchanged.
\item \link{bundle}: wrapped in a single-element \link{bundle_set}. An optional \code{name}
argument sets the element name (defaults to \code{"bundle_1"}).
}
}
\examples{
pts <- matrix(runif(15), ncol = 3, dimnames = list(NULL, c("X", "Y", "Z")))
b <- bundle(streamlines = list(streamline(points = pts)))
bs <- as_bundle_set(b, name = "sub-01")
bs@n_bundles  # 1
}
\seealso{
\code{\link[=bundle_set]{bundle_set()}}, \code{\link[=bind_bundle_sets]{bind_bundle_sets()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shape.R
\name{get_curvature}
\alias{get_curvature}
\title{Curvature of a streamline}
\usage{
get_curvature(x)
}
\arguments{
\item{x}{A \link{streamline} object.}
}
\value{
A non-negative numeric vector of length \code{x@n_points} giving the
curvature \eqn{\kappa(s)} at each sampled point along the streamline.
Higher values indicate sharper bending at that location.
}
\description{
\code{get_curvature()} is function that computes the curvature of a
\link{streamline} object. The curvature \eqn{\kappa(s)} at each point
along the arc-length abscissa is computed using cubic smoothing
splines (3 degrees of freedom per component).
}
\examples{
pts <- matrix(runif(30), ncol = 3)
colnames(pts) <- c("X", "Y", "Z")
sl <- streamline(points = pts)
get_curvature(sl)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shape.R
\name{get_sinuosity}
\alias{get_sinuosity}
\title{Sinuosity of a streamline}
\usage{
get_sinuosity(x)
}
\arguments{
\item{x}{A \link{streamline} object.}
}
\value{
A numeric scalar \eqn{\ge 1}.
}
\description{
\code{get_sinuosity()} is a function that computes the ratio of curvilinear
length to Euclidean length for a \link{streamline} object, with a value of
1 indicating a perfectly straight streamline and larger values indicating
greater curviness.
}
\examples{
pts <- matrix(runif(30), ncol = 3)
colnames(pts) <- c("X", "Y", "Z")
sl <- streamline(points = pts)
get_sinuosity(sl)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shape.R
\name{add_shape_descriptors-fiber-streamline-method}
\alias{add_shape_descriptors-fiber-streamline-method}
\alias{add_shape_descriptors,fiber::streamline-method}
\title{\code{\link[=add_shape_descriptors]{add_shape_descriptors()}} method for \code{streamline} objects}
\arguments{
\item{x}{A \link{streamline} object.}

\item{descriptors}{A character vector of shape descriptors to add. Defaults
to all available descriptors: \code{c("euclidean_length", "curvilinear_length", "sinuosity", "curvature", "torsion")}.}
}
\value{
A \link{streamline} with the specified shape descriptors added to the
\verb{@streamline_data} or \verb{@point_data} slots as appropriate.
}
\description{
Adds multiple shape descriptors to a single \link{streamline} object.
}
\seealso{
\code{\link[=add_shape_descriptors]{add_shape_descriptors()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shape.R
\name{compute_hausdorff_distance,any-method}
\alias{compute_hausdorff_distance,any-method}
\title{\code{\link[=compute_hausdorff_distance]{compute_hausdorff_distance()}} catch-all method}
\value{
Does not return a value; always throws an error for unsupported
input types.
}
\description{
\code{\link[=compute_hausdorff_distance]{compute_hausdorff_distance()}} catch-all method
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parametrize.R
\name{reparametrize}
\alias{reparametrize}
\title{Reparametrize a streamline or bundle onto a uniform arc-length grid}
\usage{
reparametrize(x, n_points = NULL)
}
\arguments{
\item{x}{A \link{streamline} or \link{bundle} object.}

\item{n_points}{Number of equally-spaced arc-length points to use.
\itemize{
\item For a single \link{streamline}, defaults to \code{nrow(x@points)}.
\item For a \link{bundle}, defaults to the rounded mean number of points across
all streamlines.
}

Pass \code{NULL} to use these defaults explicitly.}
}
\value{
An object of the same class as \code{x} reparametrized onto the new
grid.
}
\description{
\code{reparametrize()} is an S7 generic that resamples the 3-D coordinates (and
any \verb{@point_data} attributes) of a tractography object onto a uniform
arc-length grid using linear interpolation, with methods available for the
following classes:

\itemize{
\item \code{\link[=reparametrize-fiber-bundle-method]{fiber::bundle}}
\item \code{\link[=reparametrize-fiber-streamline-method]{fiber::streamline}}
}
}
\examples{
# reparametrize a single streamline to 10 points
pts <- matrix(runif(30), ncol = 3)
colnames(pts) <- c("X", "Y", "Z")
sl <- streamline(points = pts)
sl_reparam <- reparametrize(sl, n_points = 10)
# reparametrize a bundle to the mean number of points across its streamlines
sl1 <- streamline(points = pts)
pts2 <- matrix(runif(60), ncol = 3)
colnames(pts2) <- c("X", "Y", "Z")
sl2 <- streamline(points = pts2)
b <- bundle(streamlines = list(sl1, sl2))
bundle_reparam <- reparametrize(b)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shape.R
\name{get_euclidean_length}
\alias{get_euclidean_length}
\title{Euclidean length of a streamline}
\usage{
get_euclidean_length(x)
}
\arguments{
\item{x}{A \link{streamline} object.}
}
\value{
A non-negative numeric scalar.
}
\description{
\code{get_euclidean_length()} is a function that computes the Euclidean
(straight-line) distance of a \link{streamline} object.
}
\examples{
pts <- matrix(runif(30), ncol = 3)
colnames(pts) <- c("X", "Y", "Z")
sl <- streamline(points = pts)
get_euclidean_length(sl)
}
